# -*- coding: utf-8 -*-
# Copyright: (c) 2017 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

from itertools import product

import pytest

# the module we are actually testing (sort of)
from ansible.module_utils.facts.system.distribution import DistributionFactCollector


# to generate the testcase data, you can use the script gen_distribution_version_testcase.py in hacking/tests
TESTSETS = [
    {
        "platform.dist": [
            "centos",
            "7.2.1511",
            "Core"
        ],
        "input": {
            "/etc/redhat-release": "CentOS Linux release 7.2.1511 (Core) \n",
            "/etc/os-release": (
                "NAME=\"CentOS Linux\"\nVERSION=\"7 (Core)\"\nID=\"centos\"\nID_LIKE=\"rhel fedora\"\nVERSION_ID=\"7\"\n"
                "PRETTY_NAME=\"CentOS Linux 7 (Core)\"\nANSI_COLOR=\"0;31\"\nCPE_NAME=\"cpe:/o:centos:centos:7\"\n"
                "HOME_URL=\"https://www.centos.org/\"\nBUG_REPORT_URL=\"https://bugs.centos.org/\"\n\nCENTOS_MANTISBT_PROJECT=\"CentOS-7\"\n"
                "CENTOS_MANTISBT_PROJECT_VERSION=\"7\"\nREDHAT_SUPPORT_PRODUCT=\"centos\"\nREDHAT_SUPPORT_PRODUCT_VERSION=\"7\"\n\n"
            ),
            "/etc/system-release": "CentOS Linux release 7.2.1511 (Core) \n"
        },
        "name": "CentOS 7.2.1511",
        "result": {
            "distribution_release": "Core",
            "distribution": "CentOS",
            "distribution_major_version": "7",
            "os_family": "RedHat",
            "distribution_version": "7.2.1511",
        }
    },
    {
        "name": "CentOS 6.7",
        "platform.dist": [
            "centos",
            "6.7",
            "Final"
        ],
        "input": {
            "/etc/redhat-release": "CentOS release 6.7 (Final)\n",
            "/etc/lsb-release": (
                "LSB_VERSION=base-4.0-amd64:base-4.0-noarch:core-4.0-amd64:core-4.0-noarch:graphics-4.0-amd64:graphics-4.0-noarch:"
                "printing-4.0-amd64:printing-4.0-noarch\n"
            ),
            "/etc/system-release": "CentOS release 6.7 (Final)\n"
        },
        "result": {
            "distribution_release": "Final",
            "distribution": "CentOS",
            "distribution_major_version": "6",
            "os_family": "RedHat",
            "distribution_version": "6.7"
        }
    },
    {
        "name": "RedHat 7.2",
        "platform.dist": [
            "redhat",
            "7.2",
            "Maipo"
        ],
        "input": {
            "/etc/redhat-release": "Red Hat Enterprise Linux Server release 7.2 (Maipo)\n",
            "/etc/os-release": (
                "NAME=\"Red Hat Enterprise Linux Server\"\nVERSION=\"7.2 (Maipo)\"\nID=\"rhel\"\nID_LIKE=\"fedora\"\nVERSION_ID=\"7.2\"\n"
                "PRETTY_NAME=\"Red Hat Enterprise Linux Server 7.2 (Maipo)\"\nANSI_COLOR=\"0;31\"\n"
                "CPE_NAME=\"cpe:/o:redhat:enterprise_linux:7.2:GA:server\"\nHOME_URL=\"https://www.redhat.com/\"\n"
                "BUG_REPORT_URL=\"https://bugzilla.redhat.com/\"\n\nREDHAT_BUGZILLA_PRODUCT=\"Red Hat Enterprise Linux 7\"\n"
                "REDHAT_BUGZILLA_PRODUCT_VERSION=7.2\nREDHAT_SUPPORT_PRODUCT=\"Red Hat Enterprise Linux\"\n"
                "REDHAT_SUPPORT_PRODUCT_VERSION=\"7.2\"\n"
            ),
            "/etc/system-release": "Red Hat Enterprise Linux Server release 7.2 (Maipo)\n"
        },
        "result": {
            "distribution_release": "Maipo",
            "distribution": "RedHat",
            "distribution_major_version": "7",
            "os_family": "RedHat",
            "distribution_version": "7.2"
        }
    },
    {
        "name": "RedHat 6.7",
        "platform.dist": [
            "redhat",
            "6.7",
            "Santiago"
        ],
        "input": {
            "/etc/redhat-release": "Red Hat Enterprise Linux Server release 6.7 (Santiago)\n",
            "/etc/lsb-release": (
                "LSB_VERSION=base-4.0-amd64:base-4.0-noarch:core-4.0-amd64:core-4.0-noarch:graphics-4.0-amd64:graphics-4.0-noarch:"
                "printing-4.0-amd64:printing-4.0-noarch\n"
            ),
            "/etc/system-release": "Red Hat Enterprise Linux Server release 6.7 (Santiago)\n"
        },
        "result": {
            "distribution_release": "Santiago",
            "distribution": "RedHat",
            "distribution_major_version": "6",
            "os_family": "RedHat",
            "distribution_version": "6.7"
        }
    },
    {
        "name": "Virtuozzo 7.3",
        "platform.dist": [
            "redhat",
            "7.3",
            ""
        ],
        "input": {
            "/etc/redhat-release": "Virtuozzo Linux release 7.3\n",
            "/etc/os-release": (
                "NAME=\"Virtuozzo\"\n"
                "VERSION=\"7.0.3\"\n"
                "ID=\"virtuozzo\"\n"
                "ID_LIKE=\"rhel fedora\"\n"
                "VERSION_ID=\"7\"\n"
                "PRETTY_NAME=\"Virtuozzo release 7.0.3\"\n"
                "ANSI_COLOR=\"0;31\"\n"
                "CPE_NAME=\"cpe:/o:virtuozzoproject:vz:7\"\n"
                "HOME_URL=\"http://www.virtuozzo.com\"\n"
                "BUG_REPORT_URL=\"https://bugs.openvz.org/\"\n"
            ),
            "/etc/system-release": "Virtuozzo release 7.0.3 (640)\n"
        },
        "result": {
            "distribution_release": "NA",
            "distribution": "Virtuozzo",
            "distribution_major_version": "7",
            "os_family": "RedHat",
            "distribution_version": "7.3"
        }
    },
    {
        "name": "openSUSE Leap 42.1",
        "input": {
            "/etc/os-release": """
NAME="openSUSE Leap"
VERSION="42.1"
VERSION_ID="42.1"
PRETTY_NAME="openSUSE Leap 42.1 (x86_64)"
ID=opensuse
ANSI_COLOR="0;32"
CPE_NAME="cpe:/o:opensuse:opensuse:42.1"
BUG_REPORT_URL="https://bugs.opensuse.org"
HOME_URL="https://opensuse.org/"
ID_LIKE="suse"
""",
            "/etc/SuSE-release": """
openSUSE 42.1 (x86_64)
VERSION = 42.1
CODENAME = Malachite
# /etc/SuSE-release is deprecated and will be removed in the future, use /etc/os-release instead
"""
        },
        "platform.dist": ['SuSE', '42.1', 'x86_64'],
        "result": {
            "distribution": "openSUSE Leap",
            "distribution_major_version": "42",
            "distribution_release": "1",
            "os_family": "Suse",
            "distribution_version": "42.1",
        }
    },
    {
        'name': 'openSUSE 13.2',
        'input': {
            '/etc/SuSE-release': """openSUSE 13.2 (x86_64)
VERSION = 13.2
CODENAME = Harlequin
# /etc/SuSE-release is deprecated and will be removed in the future, use /etc/os-release instead
""",
            '/etc/os-release': """NAME=openSUSE
VERSION="13.2 (Harlequin)"
VERSION_ID="13.2"
PRETTY_NAME="openSUSE 13.2 (Harlequin) (x86_64)"
ID=opensuse
ANSI_COLOR="0;32"
CPE_NAME="cpe:/o:opensuse:opensuse:13.2"
BUG_REPORT_URL="https://bugs.opensuse.org"
HOME_URL="https://opensuse.org/"
ID_LIKE="suse"
"""
        },
        'platform.dist': ('SuSE', '13.2', 'x86_64'),
        'result': {
            'distribution': u'openSUSE',
            'distribution_major_version': u'13',
            'distribution_release': u'2',
            'os_family': u'Suse',
            'distribution_version': u'13.2'
        }
    },
    {
        "platform.dist": [
            "",
            "",
            ""
        ],
        "input": {
            "/etc/os-release": (
                "NAME=\"openSUSE Tumbleweed\"\n# VERSION=\"20160917\"\nID=opensuse\nID_LIKE=\"suse\"\nVERSION_ID=\"20160917\"\n"
                "PRETTY_NAME=\"openSUSE Tumbleweed\"\nANSI_COLOR=\"0;32\"\nCPE_NAME=\"cpe:/o:opensuse:tumbleweed:20160917\"\n"
                "BUG_REPORT_URL=\"https://bugs.opensuse.org\"\nHOME_URL=\"https://www.opensuse.org/\"\n"
            )
        },
        "name": "openSUSE Tumbleweed 20160917",
        "result": {
            "distribution_release": "",
            "distribution": "openSUSE Tumbleweed",
            "distribution_major_version": "20160917",
            "os_family": "Suse",
            "distribution_version": "20160917"
        }
    },
    {
        "platform.dist": [
            "",
            "",
            ""
        ],
        "input": {
            "/etc/os-release": (
                "NAME=\"openSUSE Leap\"\n# VERSION=\"15.0\"\nID=opensuse-leap\nID_LIKE=\"suse opensuse\"\nVERSION_ID=\"15.0\"\n"
                "PRETTY_NAME=\"openSUSE Leap 15.0\"\nANSI_COLOR=\"0;32\"\nCPE_NAME=\"cpe:/o:opensuse:leap:15.0\"\n"
                "BUG_REPORT_URL=\"https://bugs.opensuse.org\"\nHOME_URL=\"https://www.opensuse.org/\"\n"
            )
        },
        "name": "openSUSE Leap 15.0",
        "result": {
            "distribution_release": "0",
            "distribution": "openSUSE Leap",
            "distribution_major_version": "15",
            "os_family": "Suse",
            "distribution_version": "15.0"
        }
    },
    {  # see https://github.com/ansible/ansible/issues/14837
        "name": "SLES 11.3",
        "input": {
            "/etc/SuSE-release": """
SUSE Linux Enterprise Server 11 (x86_64)
VERSION = 11
PATCHLEVEL = 3
"""
        },
        "platform.dist": ['SuSE', '11', 'x86_64'],
        "result": {
            "distribution": "SLES",
            "distribution_major_version": "11",
            "distribution_release": "3",
            "os_family": "Suse",
            "distribution_version": "11.3",
        }
    },
    {  # see https://github.com/ansible/ansible/issues/14837
        "name": "SLES 11.4",
        "input": {
            "/etc/SuSE-release": """
SUSE Linux Enterprise Server 11 (x86_64)
VERSION = 11
PATCHLEVEL = 4
""",
            "/etc/os-release": """
NAME="SLES"
VERSION="11.4"
VERSION_ID="11.4"
PRETTY_NAME="SUSE Linux Enterprise Server 11 SP4"
ID="sles"
ANSI_COLOR="0;32"
CPE_NAME="cpe:/o:suse:sles:11:4"
""",
        },
        "platform.dist": ['SuSE', '11', 'x86_64'],
        "result":{
            "distribution": "SLES",
            "distribution_major_version": "11",
            "distribution_release": "4",
            "os_family": "Suse",
            "distribution_version": "11.4",
        }
    },
    {  # see https://github.com/ansible/ansible/issues/14837
        "name": "SLES 12 SP0",
        "input": {
            "/etc/SuSE-release": """
SUSE Linux Enterprise Server 12 (x86_64)
VERSION = 12
PATCHLEVEL = 0
# This file is deprecated and will be removed in a future service pack or release.
# Please check /etc/os-release for details about this release.
""",
            "/etc/os-release": """
NAME="SLES"
VERSION="12"
VERSION_ID="12"
PRETTY_NAME="SUSE Linux Enterprise Server 12"
ID="sles"
ANSI_COLOR="0;32"
CPE_NAME="cpe:/o:suse:sles:12"
""",
        },
        "platform.dist": ['SuSE', '12', 'x86_64'],
        "result": {
            "distribution": "SLES",
            "distribution_major_version": "12",
            "distribution_release": "0",
            "os_family": "Suse",
            "distribution_version": "12",
        }
    },
    {  # see https://github.com/ansible/ansible/issues/14837
        "name": "SLES 12 SP1",
        "input": {
            "/etc/SuSE-release": """
SUSE Linux Enterprise Server 12 (x86_64)
VERSION = 12
PATCHLEVEL = 0
# This file is deprecated and will be removed in a future service pack or release.
# Please check /etc/os-release for details about this release.
""",
            "/etc/os-release": """
NAME="SLES"
VERSION="12-SP1"
VERSION_ID="12.1"
PRETTY_NAME="SUSE Linux Enterprise Server 12 SP1"
ID="sles"
ANSI_COLOR="0;32"
CPE_NAME="cpe:/o:suse:sles:12:sp1"
            """,
        },
        "platform.dist": ['SuSE', '12', 'x86_64'],
        "result":{
            "distribution": "SLES",
            "distribution_major_version": "12",
            "distribution_release": "1",
            "os_family": "Suse",
            "distribution_version": "12.1",
        }
    },

    {
        "name": "Debian stretch/sid",
        "input": {
            "/etc/os-release": """
PRETTY_NAME="Debian GNU/Linux stretch/sid"
NAME="Debian GNU/Linux"
ID=debian
HOME_URL="https://www.debian.org/"
SUPPORT_URL="https://www.debian.org/support"
BUG_REPORT_URL="https://bugs.debian.org/"
""",
            "/etc/debian_version": """
            stretch/sid
            """,
        },
        "platform.dist": ('debian', 'stretch/sid', ''),
        "result": {
            "distribution": "Debian",
            "distribution_major_version": "stretch/sid",
            "distribution_release": "NA",
            "os_family": "Debian",
            "distribution_version": "stretch/sid",
        }
    },
    {
        'name': "Debian 7.9",
        'input': {
            '/etc/os-release': """PRETTY_NAME="Debian GNU/Linux 7 (wheezy)"
NAME="Debian GNU/Linux"
VERSION_ID="7"
VERSION="7 (wheezy)"
ID=debian
ANSI_COLOR="1;31"
HOME_URL="http://www.debian.org/"
SUPPORT_URL="http://www.debian.org/support/"
BUG_REPORT_URL="http://bugs.debian.org/"
"""
        },
        'platform.dist': ('debian', '7.9', ''),
        'result': {
            'distribution': u'Debian',
            'distribution_major_version': u'7',
            'distribution_release': u'wheezy',
            "os_family": "Debian",
            'distribution_version': u'7.9'
        }
    },
    {
        'name': "SteamOS 2.0",
        'input': {
            '/etc/os-release': """PRETTY_NAME="SteamOS GNU/Linux 2.0 (brewmaster)"
NAME="SteamOS GNU/Linux"
VERSION_ID="2"
VERSION="2 (brewmaster)"
ID=steamos
ID_LIKE=debian
HOME_URL="http://www.steampowered.com/"
SUPPORT_URL="http://support.steampowered.com/"
BUG_REPORT_URL="http://support.steampowered.com/"
""",
            '/etc/lsb-release': """DISTRIB_ID=SteamOS
DISTRIB_RELEASE=2.0
DISTRIB_CODENAME=brewmaster
DISTRIB_DESCRIPTION="SteamOS 2.0"
"""
        },
        'platform.dist': ('Steamos', '2.0', 'brewmaster'),
        'result': {
            'distribution': u'SteamOS',
            'distribution_major_version': u'2',
            'distribution_release': u'brewmaster',
            "os_family": "Debian",
            'distribution_version': u'2.0'
        }
    },
    {
        'name': "Devuan",
        'input': {
            '/etc/os-release': """PRETTY_NAME="Devuan GNU/Linux 1 (jessie)"
NAME="Devuan GNU/Linux"
VERSION_ID="1"
VERSION="1 (jessie)"
ID=devuan
HOME_URL="http://www.devuan.org/"
SUPPORT_URL="http://www.devuan.org/support/"
BUG_REPORT_URL="https://bugs.devuan.org/"
"""
        },
        'platform.dist': ('', '', ''),
        'result': {
            'distribution': u'Devuan',
            'distribution_major_version': u'1',
            'distribution_release': u'jessie',
            'os_family': 'Debian',
            'distribution_version': u'1'
        }
    },
    {
        "platform.dist": [
            "Ubuntu",
            "16.04",
            "xenial"
        ],
        "input": {
            "/etc/os-release": (
                "NAME=\"Ubuntu\"\nVERSION=\"16.04 LTS (Xenial Xerus)\"\nID=ubuntu\nID_LIKE=debian\nPRETTY_NAME=\"Ubuntu 16.04 LTS\"\n"
                "VERSION_ID=\"16.04\"\nHOME_URL=\"http://www.ubuntu.com/\"\nSUPPORT_URL=\"http://help.ubuntu.com/\"\n"
                "BUG_REPORT_URL=\"http://bugs.launchpad.net/ubuntu/\"\nUBUNTU_CODENAME=xenial\n"
            ),
            "/etc/lsb-release": "DISTRIB_ID=Ubuntu\nDISTRIB_RELEASE=16.04\nDISTRIB_CODENAME=xenial\nDISTRIB_DESCRIPTION=\"Ubuntu 16.04 LTS\"\n"
        },
        "name": "Ubuntu 16.04",
        "result": {
            "distribution_release": "xenial",
            "distribution": "Ubuntu",
            "distribution_major_version": "16",
            "os_family": "Debian",
            "distribution_version": "16.04"
        }
    },
    {
        'name': "Ubuntu 10.04 guess",
        'input':
            {
                '/etc/lsb-release': """DISTRIB_ID=Ubuntu
DISTRIB_RELEASE=10.04
DISTRIB_CODENAME=lucid
DISTRIB_DESCRIPTION="Ubuntu 10.04.4 LTS
"""
            },
        'platform.dist': ('Ubuntu', '10.04', 'lucid'),
        'result':
            {
                'distribution': u'Ubuntu',
                'distribution_major_version': u'10',
                'distribution_release': u'lucid',
                "os_family": "Debian",
                'distribution_version': u'10.04'
            }
    },
    {
        'name': "Ubuntu 14.04",
        'input': {
            '/etc/lsb-release': """DISTRIB_ID=Ubuntu
DISTRIB_RELEASE=14.04
DISTRIB_CODENAME=trusty
DISTRIB_DESCRIPTION="Ubuntu 14.04.4 LTS"
""",
            '/etc/os-release': """NAME="Ubuntu"
VERSION="14.04.4 LTS, Trusty Tahr"
ID=ubuntu
ID_LIKE=debian
PRETTY_NAME="Ubuntu 14.04.4 LTS"
VERSION_ID="14.04"
HOME_URL="http://www.ubuntu.com/"
SUPPORT_URL="http://help.ubuntu.com/"
BUG_REPORT_URL="http://bugs.launchpad.net/ubuntu/"
"""
        },
        'platform.dist': ('Ubuntu', '14.04', 'trusty'),
        'result': {
            'distribution': u'Ubuntu',
            'distribution_major_version': u'14',
            'distribution_release': u'trusty',
            "os_family": "Debian",
            'distribution_version': u'14.04'
        }
    },
    {
        'name': "Ubuntu 12.04",
        'input': {'/etc/lsb-release': """DISTRIB_ID=Ubuntu
DISTRIB_RELEASE=12.04
DISTRIB_CODENAME=precise
DISTRIB_DESCRIPTION="Ubuntu 12.04.5 LTS"
""",
                  '/etc/os-release': """NAME="Ubuntu"
VERSION="12.04.5 LTS, Precise Pangolin"
ID=ubuntu
ID_LIKE=debian
PRETTY_NAME="Ubuntu precise (12.04.5 LTS)"
VERSION_ID="12.04"
"""},
        'platform.dist': ('Ubuntu', '12.04', 'precise'),
        'result': {'distribution': u'Ubuntu',
                   'distribution_major_version': u'12',
                   'distribution_release': u'precise',
                   "os_family": "Debian",
                   'distribution_version': u'12.04'}
    },
    {
        "platform.dist": [
            "neon",
            "16.04",
            "xenial"
        ],
        "input": {
            "/etc/os-release": ("NAME=\"KDE neon\"\nVERSION=\"5.8\"\nID=neon\nID_LIKE=\"ubuntu debian\"\nPRETTY_NAME=\"KDE neon User Edition 5.8\"\n"
                                "VERSION_ID=\"16.04\"\nHOME_URL=\"http://neon.kde.org/\"\nSUPPORT_URL=\"http://neon.kde.org/\"\n"
                                "BUG_REPORT_URL=\"http://bugs.kde.org/\"\nVERSION_CODENAME=xenial\nUBUNTU_CODENAME=xenial\n"),
            "/etc/lsb-release": "DISTRIB_ID=neon\nDISTRIB_RELEASE=16.04\nDISTRIB_CODENAME=xenial\nDISTRIB_DESCRIPTION=\"KDE neon User Edition 5.8\"\n"
        },
        "name": "KDE neon 16.04",
        "result": {
            "distribution_release": "xenial",
            "distribution": "KDE neon",
            "distribution_major_version": "16",
            "os_family": "Debian",
            "distribution_version": "16.04"
        }
    },
    {
        'name': 'Core OS',
        'input': {
            '/etc/os-release': """
NAME=CoreOS
ID=coreos
VERSION=976.0.0
VERSION_ID=976.0.0
BUILD_ID=2016-03-03-2324
PRETTY_NAME="CoreOS 976.0.0 (Coeur Rouge)"
ANSI_COLOR="1;32"
HOME_URL="https://coreos.com/"
BUG_REPORT_URL="https://github.com/coreos/bugs/issues"
""",
            '/etc/lsb-release': """DISTRIB_ID=CoreOS
DISTRIB_RELEASE=976.0.0
DISTRIB_CODENAME="Coeur Rouge"
DISTRIB_DESCRIPTION="CoreOS 976.0.0 (Coeur Rouge)"
""",
        },
        'platform.dist': ('', '', ''),
        'platform.release': '',
        'result': {
            "distribution": "CoreOS",
            "distribution_major_version": "NA",
            "distribution_release": "NA",
            "distribution_version": "976.0.0",
        }
    },
    # Solaris and derivatives: https://gist.github.com/natefoo/7af6f3d47bb008669467
    {
        "name": "SmartOS Global Zone",
        "uname_v": "joyent_20160330T234717Z",
        "result": {
            "distribution_release": "SmartOS 20160330T234717Z x86_64",
            "distribution": "SmartOS",
            "os_family": "Solaris",
            "distribution_version": "joyent_20160330T234717Z"
        },
        "platform.dist": [
            "",
            "",
            ""
        ],
        "input": {
            "/etc/release": ("                       SmartOS 20160330T234717Z x86_64\n"
                             "              Copyright 2010 Sun Microsystems, Inc.  All Rights Reserved.\n"
                             "              Copyright 2010-2012 Joyent, Inc.  All Rights Reserved.\n"
                             "                        Use is subject to license terms.\n\n"
                             "   Built with the following components:\n\n[\n"
                             "        { \"repo\": \"smartos-live\", \"branch\": \"release-20160331\", \"rev\": \"a77c410f2afe6dc9853a915733caec3609cc50f1\", "
                             "\"commit_date\": \"1459340323\", \"url\": \"git@github.com:joyent/smartos-live.git\" }\n        , "
                             "{ \"repo\": \"illumos-joyent\", \"branch\": \"release-20160331\", \"rev\": \"ab664c06caf06e9ce7586bff956e7709df1e702e\", "
                             "\"commit_date\": \"1459362533\", \"url\": \"/root/data/jenkins/workspace/smartos/MG/build/illumos-joyent\" }\n"
                             "        , { \"repo\": \"illumos-extra\", \"branch\": \"release-20160331\", "
                             "\"rev\": \"cc723855bceace3df7860b607c9e3827d47e0ff4\", \"commit_date\": \"1458153188\", "
                             "\"url\": \"/root/data/jenkins/workspace/smartos/MG/build/illumos-extra\" }\n        , "
                             "{ \"repo\": \"kvm\", \"branch\": \"release-20160331\", \"rev\": \"a8befd521c7e673749c64f118585814009fe4b73\", "
                             "\"commit_date\": \"1450081968\", \"url\": \"/root/data/jenkins/workspace/smartos/MG/build/illumos-kvm\" }\n        , "
                             "{ \"repo\": \"kvm-cmd\", \"branch\": \"release-20160331\", \"rev\": \"c1a197c8e4582c68739ab08f7e3198b2392c9820\", "
                             "\"commit_date\": \"1454723558\", \"url\": \"/root/data/jenkins/workspace/smartos/MG/build/illumos-kvm-cmd\" }\n        , "
                             "{ \"repo\": \"mdata-client\", \"branch\": \"release-20160331\", \"rev\": \"58158c44603a3316928975deccc5d10864832770\", "
                             "\"commit_date\": \"1429917227\", \"url\": \"/root/data/jenkins/workspace/smartos/MG/build/mdata-client\" }\n]\n")
        },
        "platform.system": "SunOS"
    },
    {
        "name": "SmartOS Zone",
        "uname_v": "joyent_20160330T234717Z",
        "result": {
            "distribution_release": "SmartOS x86_64",
            "distribution": "SmartOS",
            "os_family": "Solaris",
            "distribution_version": "14.3.0"
        },
        "platform.dist": [
            "",
            "",
            ""
        ],
        "input": {
            "/etc/release": ("                                SmartOS x86_64\n              Copyright 2010 Sun Microsystems, Inc.  All Rights Reserved.\n"
                             "              Copyright 2010-2013 Joyent, Inc.  All Rights Reserved.\n                        Use is subject to license terms.\n"
                             "                   See joyent_20141002T182809Z for assembly date and time.\n"),
            "/etc/product": "Name: Joyent Instance\nImage: base64 14.3.0\nDocumentation: http://wiki.joyent.com/jpc2/Base+Instance\n"
        },
        "platform.system": "SunOS"
    },
    {
        "name": "OpenIndiana",
        "uname_v": "oi_151a9",
        "result": {
            "distribution_release": "OpenIndiana Development oi_151.1.9 X86 (powered by illumos)",
            "distribution": "OpenIndiana",
            "os_family": "Solaris",
            "distribution_version": "oi_151a9"
        },
        "platform.dist": [
            "",
            "",
            ""
        ],
        "input": {
            "/etc/release": ("             OpenIndiana Development oi_151.1.9 X86 (powered by illumos)\n        Copyright 2011 Oracle and/or its affiliates. "
                             "All rights reserved.\n                        Use is subject to license terms.\n                           "
                             "Assembled 17 January 2014\n")
        },
        "platform.system": "SunOS"
    },
    {
        "name": "OmniOS",
        "uname_v": "omnios-10b9c79",
        "result": {
            "distribution_release": "OmniOS v11 r151012",
            "distribution": "OmniOS",
            "os_family": "Solaris",
            "distribution_version": "r151012"
        },
        "platform.dist": [
            "",
            "",
            ""
        ],
        #        "platform.release": 'OmniOS',
        "input": {
            "/etc/release": (
                "  OmniOS v11 r151012\n  Copyright 2014 OmniTI Computer Consulting, Inc. All rights reserved.\n  Use is subject to license terms.\n\n"
            )
        },
        "platform.system": "SunOS"
    },
    {
        "name": "Nexenta 3",
        "uname_v": "NexentaOS_134f",
        "result": {
            "distribution_release": "Open Storage Appliance v3.1.6",
            "distribution": "Nexenta",
            "os_family": "Solaris",
            "distribution_version": "3.1.6"
        },
        "platform.dist": [
            "",
            "",
            ""
        ],
        "platform.release:": "",
        "input": {
            "/etc/release": ("                         Open Storage Appliance v3.1.6\n           Copyright (c) 2014 Nexenta Systems, Inc.  "
                             "All Rights Reserved.\n           Copyright (c) 2011 Oracle.  All Rights Reserved.\n                         "
                             "Use is subject to license terms.\n")
        },
        "platform.system": "SunOS"
    },
    {
        "name": "Nexenta 4",
        "uname_v": "NexentaOS_4:cd604cd066",
        "result": {
            "distribution_release": "Open Storage Appliance 4.0.3-FP2",
            "distribution": "Nexenta",
            "os_family": "Solaris",
            "distribution_version": "4.0.3-FP2"
        },
        "platform.dist": [
            "",
            "",
            ""
        ],
        "input": {
            "/etc/release": ("                        Open Storage Appliance 4.0.3-FP2\n           Copyright (c) 2014 Nexenta Systems, Inc.  "
                             "All Rights Reserved.\n           Copyright (c) 2010 Oracle.  All Rights Reserved.\n                        "
                             "Use is subject to license terms.\n")
        },
        "platform.system": "SunOS"
    },
    {
        "name": "Solaris 10",
        "uname_v": "Generic_141445-09",
        "result": {
            "distribution_release": "Solaris 10 10/09 s10x_u8wos_08a X86",
            "distribution": "Solaris",
            "os_family": "Solaris",
            "distribution_version": "10"
        },
        "platform.dist": [
            "",
            "",
            ""
        ],
        "input": {
            "/etc/release": ("                       Solaris 10 10/09 s10x_u8wos_08a X86\n           Copyright 2009 Sun Microsystems, Inc.  "
                             "All Rights Reserved.\n                        Use is subject to license terms.\n                           "
                             "Assembled 16 September 2009\n")
        },
        "platform.system": "SunOS"
    },
    {
        "name": "Solaris 11",
        "uname_v": "11.0",
        "result": {
            "distribution_release": "Oracle Solaris 11 11/11 X86",
            "distribution": "Solaris",
            "os_family": "Solaris",
            "distribution_version": "11"
        },
        "platform.dist": [
            "",
            "",
            ""
        ],
        "input": {
            "/etc/release": ("                           Oracle Solaris 11 11/11 X86\n  Copyright (c) 1983, 2011, Oracle and/or its affiliates.  "
                             "All rights reserved.\n                            Assembled 18 October 2011\n")
        },
        "platform.system": "SunOS"
    },
    {
        "name": "Solaris 11.3",
        "platform.dist": [
            "",
            "",
            ""
        ],
        "input": {
            "/etc/release": (
                "                             Oracle Solaris 11.3 X86\n  Copyright (c) 1983, 2015, Oracle and/or its affiliates.  "
                "All rights reserved.\n                            Assembled 06 October 2015\n"
            )
        },
        "platform.system": "SunOS",
        "result": {
            "distribution_release": "Oracle Solaris 11.3 X86",
            "distribution": "Solaris",
            "os_family": "Solaris",
            "distribution_version": "11.3"
        }
    },
    {
        "name": "Solaris 10",
        "platform.dist": [
            "",
            "",
            ""
        ],
        "input": {
            "/etc/release": ("                    Oracle Solaris 10 1/13 s10x_u11wos_24a X86\n  Copyright (c) 1983, 2013, Oracle and/or its affiliates. "
                             "All rights reserved.\n                            Assembled 17 January 2013\n")
        },
        "platform.system": "SunOS",
        "result": {
            "distribution_release": "Oracle Solaris 10 1/13 s10x_u11wos_24a X86",
            "distribution": "Solaris",
            "os_family": "Solaris",
            "distribution_version": "10"
        }
    },
    {
        "name": "Fedora 22",
        "platform.dist": [
            "fedora",
            "22",
            "Twenty Two"
        ],
        "input": {
            "/etc/redhat-release": "Fedora release 22 (Twenty Two)\n",
            "/etc/os-release": (
                "NAME=Fedora\nVERSION=\"22 (Twenty Two)\"\nID=fedora\nVERSION_ID=22\nPRETTY_NAME=\"Fedora 22 (Twenty Two)\"\n"
                "ANSI_COLOR=\"0;34\"\nCPE_NAME=\"cpe:/o:fedoraproject:fedora:22\"\nHOME_URL=\"https://fedoraproject.org/\"\n"
                "BUG_REPORT_URL=\"https://bugzilla.redhat.com/\"\nREDHAT_BUGZILLA_PRODUCT=\"Fedora\"\nREDHAT_BUGZILLA_PRODUCT_VERSION=22\n"
                "REDHAT_SUPPORT_PRODUCT=\"Fedora\"\nREDHAT_SUPPORT_PRODUCT_VERSION=22\n"
                "PRIVACY_POLICY_URL=https://fedoraproject.org/wiki/Legal:PrivacyPolicy\n"
            ),
            "/etc/system-release": "Fedora release 22 (Twenty Two)\n"
        },
        "result": {
            "distribution_release": "Twenty Two",
            "distribution": "Fedora",
            "distribution_major_version": "22",
            "os_family": "RedHat",
            "distribution_version": "22"
        }
    },
    {
        "platform.dist": [
            "fedora",
            "25",
            "Rawhide"
        ],
        "input": {
            "/etc/redhat-release": "Fedora release 25 (Rawhide)\n",
            "/etc/os-release": (
                "NAME=Fedora\nVERSION=\"25 (Workstation Edition)\"\nID=fedora\nVERSION_ID=25\n"
                "PRETTY_NAME=\"Fedora 25 (Workstation Edition)\"\nANSI_COLOR=\"0;34\"\nCPE_NAME=\"cpe:/o:fedoraproject:fedora:25\"\n"
                "HOME_URL=\"https://fedoraproject.org/\"\nBUG_REPORT_URL=\"https://bugzilla.redhat.com/\"\n"
                "REDHAT_BUGZILLA_PRODUCT=\"Fedora\"\nREDHAT_BUGZILLA_PRODUCT_VERSION=rawhide\nREDHAT_SUPPORT_PRODUCT=\"Fedora\"\n"
                "REDHAT_SUPPORT_PRODUCT_VERSION=rawhide\nPRIVACY_POLICY_URL=https://fedoraproject.org/wiki/Legal:PrivacyPolicy\n"
                "VARIANT=\"Workstation Edition\"\nVARIANT_ID=workstation\n"
            ),
            "/etc/system-release": "Fedora release 25 (Rawhide)\n"
        },
        "name": "Fedora 25",
        "result": {
            "distribution_release": "Rawhide",
            "distribution": "Fedora",
            "distribution_major_version": "25",
            "os_family": "RedHat",
            "distribution_version": "25"
        }
    },
    {
        "platform.dist": [
            "",
            "",
            ""
        ],
        "input": {
            "/etc/sourcemage-release": ("Source Mage GNU/Linux x86_64-pc-linux-gnu\nInstalled from tarball using chroot image (Grimoire 0.61-rc) "
                                        "on Thu May 17 17:31:37 UTC 2012\n")
        },
        "name": "SMGL NA",
        "result": {
            "distribution_release": "NA",
            "distribution": "SMGL",
            "distribution_major_version": "NA",
            "os_family": "SMGL",
            "distribution_version": "NA"
        }
    },

    # ArchLinux with an empty /etc/arch-release and a /etc/os-release with "NAME=Arch Linux"
    {
        "platform.dist": [
            "",
            "",
            ""
        ],
        "input": {
            "/etc/os-release": "NAME=\"Arch Linux\"\nPRETTY_NAME=\"Arch Linux\"\nID=arch\nID_LIKE=archlinux\nANSI_COLOR=\"0;36\"\nHOME_URL=\"https://www.archlinux.org/\"\nSUPPORT_URL=\"https://bbs.archlinux.org/\"\nBUG_REPORT_URL=\"https://bugs.archlinux.org/\"\n\n",  # noqa
            "/etc/arch-release": "",
        },
        "name": "Arch Linux NA",
        "result": {
            "distribution_release": "NA",
            "distribution": "Archlinux",
            "distribution_major_version": "NA",
            "os_family": "Archlinux",
            "distribution_version": "NA"
        }
    },

    # ClearLinux https://github.com/ansible/ansible/issues/31501#issuecomment-340861535
{
    "platform.dist": [
        "Clear Linux OS",
        "26580",
        "clear-linux-os"
    ],
    "input": {
        "/etc/os-release": '''
NAME="Clear Linux OS"
VERSION=1
ID=clear-linux-os
ID_LIKE=clear-linux-os
VERSION_ID=26580
PRETTY_NAME="Clear Linux OS"
ANSI_COLOR="1;35"
HOME_URL="https://clearlinux.org"
SUPPORT_URL="https://clearlinux.org"
BUG_REPORT_URL="mailto:dev@lists.clearlinux.org"
PRIVACY_POLICY_URL="http://www.intel.com/privacy"
''',
        "/usr/lib/os-release": '''
NAME="Clear Linux OS"
VERSION=1
ID=clear-linux-os
ID_LIKE=clear-linux-os
VERSION_ID=26580
PRETTY_NAME="Clear Linux OS"
ANSI_COLOR="1;35"
HOME_URL="https://clearlinux.org"
SUPPORT_URL="https://clearlinux.org"
BUG_REPORT_URL="mailto:dev@lists.clearlinux.org"
PRIVACY_POLICY_URL="http://www.intel.com/privacy"
'''
    },
    "name": "ClearLinux 26580",
    "result": {
        "distribution_release": "clear-linux-os",
        "distribution": "Clear Linux OS",
        "distribution_major_version": "26580",
        "os_family": "ClearLinux",
        "distribution_version": "26580"
    }
},
    # ArchLinux with no /etc/arch-release but with a /etc/os-release with NAME=Arch Linux
    # The fact needs to map 'Arch Linux' to 'Archlinux' for compat with 2.3 and earlier facts
    {
        "platform.dist": [
            "",
            "",
            ""
        ],
        "input": {
            "/etc/os-release": "NAME=\"Arch Linux\"\nPRETTY_NAME=\"Arch Linux\"\nID=arch\nID_LIKE=archlinux\nANSI_COLOR=\"0;36\"\nHOME_URL=\"https://www.archlinux.org/\"\nSUPPORT_URL=\"https://bbs.archlinux.org/\"\nBUG_REPORT_URL=\"https://bugs.archlinux.org/\"\n\n",  # noqa
        },
        "name": "Arch Linux no arch-release NA",
        "result": {
            "distribution_release": "NA",
            "distribution": "Archlinux",
            "distribution_major_version": "NA",
            "os_family": "Archlinux",
            "distribution_version": "NA"
        }
    }
]


@pytest.mark.parametrize("stdin, testcase", product([{}], TESTSETS), ids=lambda x: x.get('name'), indirect=['stdin'])
def test_distribution_version(am, mocker, testcase):
    """tests the distribution parsing code of the Facts class

    testsets have
    * a name (for output/debugging only)
    * input files that are faked
      * those should be complete and also include "irrelevant" files that might be mistaken as coming from other distributions
      * all files that are not listed here are assumed to not exist at all
    * the output of pythons platform.dist()
    * results for the ansible variables distribution* and os_family

    """

    # prepare some mock functions to get the testdata in
    def mock_get_file_content(fname, default=None, strip=True):
        """give fake content if it exists, otherwise pretend the file is empty"""
        data = default
        if fname in testcase['input']:
            # for debugging
            print('faked %s for %s' % (fname, testcase['name']))
            data = testcase['input'][fname].strip()
        if strip and data is not None:
            data = data.strip()
        return data

    def mock_get_uname_version(am):
        return testcase.get('uname_v', None)

    def mock_file_exists(fname, allow_empty=False):
        if fname not in testcase['input']:
            return False

        if allow_empty:
            return True
        return bool(len(testcase['input'][fname]))

    def mock_platform_system():
        return testcase.get('platform.system', 'Linux')

    def mock_platform_release():
        return testcase.get('platform.release', '')

    def mock_platform_version():
        return testcase.get('platform.version', '')

    mocker.patch('ansible.module_utils.facts.system.distribution.get_file_content', mock_get_file_content)
    mocker.patch('ansible.module_utils.facts.system.distribution.get_uname_version', mock_get_uname_version)
    mocker.patch('ansible.module_utils.facts.system.distribution._file_exists', mock_file_exists)
    mocker.patch('platform.dist', lambda: testcase['platform.dist'])
    mocker.patch('platform.system', mock_platform_system)
    mocker.patch('platform.release', mock_platform_release)
    mocker.patch('platform.version', mock_platform_version)

    # run Facts()
    distro_collector = DistributionFactCollector()
    generated_facts = distro_collector.collect(am)

    # compare with the expected output

    # testcase['result'] has a list of variables and values it expects Facts() to set
    for key, val in testcase['result'].items():
        assert key in generated_facts
        msg = 'Comparing value of %s on %s, should: %s, is: %s' %\
            (key, testcase['name'], val, generated_facts[key])
        assert generated_facts[key] == val, msg



# (c) 2016 Red Hat Inc.
# (c) 2017 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

DOCUMENTATION = """
---
author: Ansible Networking Team
connection: netconf
short_description: Provides a persistent connection using the netconf protocol
description:
  - This connection plugin provides a connection to remote devices over the
    SSH NETCONF subsystem.  This connection plugin is typically used by
    network devices for sending and receiving RPC calls over NETCONF.
  - Note this connection plugin requires ncclient to be installed on the
    local Ansible controller.
version_added: "2.3"
requirements:
  - ncclient
options:
  host:
    description:
      - Specifies the remote device FQDN or IP address to establish the SSH
        connection to.
    default: inventory_hostname
    vars:
      - name: ansible_host
  port:
    type: int
    description:
      - Specifies the port on the remote device to listening for connections
        when establishing the SSH connection.
    default: 830
    ini:
      - section: defaults
        key: remote_port
    env:
      - name: ANSIBLE_REMOTE_PORT
    vars:
      - name: ansible_port
  network_os:
    description:
      - Configures the device platform network operating system.  This value is
        used to load a device specific netconf plugin.  If this option is not
        configured, then the default netconf plugin will be used.
    vars:
      - name: ansible_network_os
  remote_user:
    description:
      - The username used to authenticate to the remote device when the SSH
        connection is first established.  If the remote_user is not specified,
        the connection will use the username of the logged in user.
      - Can be configured form the CLI via the C(--user) or C(-u) options
    ini:
      - section: defaults
        key: remote_user
    env:
      - name: ANSIBLE_REMOTE_USER
    vars:
      - name: ansible_user
  password:
    description:
      - Configures the user password used to authenticate to the remote device
        when first establishing the SSH connection.
    vars:
      - name: ansible_password
      - name: ansible_ssh_pass
  private_key_file:
    description:
      - The private SSH key or certificate file used to to authenticate to the
        remote device when first establishing the SSH connection.
    ini:
      - section: defaults
        key: private_key_file
    env:
      - name: ANSIBLE_PRIVATE_KEY_FILE
    vars:
      - name: ansible_private_key_file
  timeout:
    type: int
    description:
      - Sets the connection time for the communicating with the remote device.
        This timeout is used as the default timeout value when awaiting a
        response after issuing a call to a RPC.  If the RPC does not return in
        timeout seconds, an error is generated.
    default: 120
  host_key_auto_add:
    type: bool
    description:
      - By default, Ansible will prompt the user before adding SSH keys to the
        known hosts file.  Enabling this option, unknown host keys will
        automatically be added to the known hosts file.
      - Be sure to fully understand the security implications of enabling this
        option on production systems as it could create a security vulnerability.
    default: 'no'
    ini:
      - section: paramiko_connection
        key: host_key_auto_add
    env:
      - name: ANSIBLE_HOST_KEY_AUTO_ADD
  look_for_keys:
    default: True
    description:
      -  enables looking for ssh keys in the usual locations for ssh keys (e.g. :file:`~/.ssh/id_*`)
    env:
      - name: ANSIBLE_PARAMIKO_LOOK_FOR_KEYS
    ini:
      - section: paramiko_connection
        key: look_for_keys
    type: boolean
  host_key_checking:
    description: 'Set this to "False" if you want to avoid host key checking by the underlying tools Ansible uses to connect to the host'
    type: boolean
    default: True
    env:
      - name: ANSIBLE_HOST_KEY_CHECKING
      - name: ANSIBLE_SSH_HOST_KEY_CHECKING
      - name: ANSIBLE_NETCONF_HOST_KEY_CHECKING
    ini:
      - section: defaults
        key: host_key_checking
      - section: paramiko_connection
        key: host_key_checking
    vars:
      - name: ansible_host_key_checking
      - name: ansible_ssh_host_key_checking
      - name: ansible_netconf_host_key_checking
  persistent_connect_timeout:
    type: int
    description:
      - Configures, in seconds, the amount of time to wait when trying to
        initially establish a persistent connection.  If this value expires
        before the connection to the remote device is completed, the connection
        will fail
    default: 30
    ini:
      - section: persistent_connection
        key: connect_timeout
    env:
      - name: ANSIBLE_PERSISTENT_CONNECT_TIMEOUT
    vars:
      - name: ansible_connect_timeout
  persistent_command_timeout:
    type: int
    description:
      - Configures, in seconds, the amount of time to wait for a command to
        return from the remote device.  If this timer is exceeded before the
        command returns, the connection plugin will raise an exception and
        close
    default: 10
    ini:
      - section: persistent_connection
        key: command_timeout
    env:
      - name: ANSIBLE_PERSISTENT_COMMAND_TIMEOUT
    vars:
      - name: ansible_command_timeout
  netconf_ssh_config:
    description:
      - This variable is used to enable bastion/jump host with netconf connection. If set to
        True the bastion/jump host ssh settings should be present in ~/.ssh/config file,
        alternatively it can be set to custom ssh configuration file path to read the
        bastion/jump host settings.
    ini:
      - section: netconf_connection
        key: ssh_config
        version_added: '2.7'
    env:
      - name: ANSIBLE_NETCONF_SSH_CONFIG
    vars:
      - name: ansible_netconf_ssh_config
        version_added: '2.7'
"""

import os
import logging
import json

from ansible.errors import AnsibleConnectionFailure, AnsibleError
from ansible.module_utils._text import to_bytes, to_native, to_text
from ansible.module_utils.parsing.convert_bool import BOOLEANS_TRUE, BOOLEANS_FALSE
from ansible.plugins.loader import netconf_loader
from ansible.plugins.connection import NetworkConnectionBase

try:
    from ncclient import manager
    from ncclient.operations import RPCError
    from ncclient.transport.errors import SSHUnknownHostError
    from ncclient.xml_ import to_ele, to_xml
    HAS_NCCLIENT = True
except ImportError:
    HAS_NCCLIENT = False

try:
    from __main__ import display
except ImportError:
    from ansible.utils.display import Display
    display = Display()

logging.getLogger('ncclient').setLevel(logging.INFO)

NETWORK_OS_DEVICE_PARAM_MAP = {
    "nxos": "nexus",
    "ios": "default",
    "sros": "alu",
    "ce": "huawei"
}


class Connection(NetworkConnectionBase):
    """NetConf connections"""

    transport = 'netconf'
    has_pipelining = False

    def __init__(self, play_context, new_stdin, *args, **kwargs):
        super(Connection, self).__init__(play_context, new_stdin, *args, **kwargs)

        self._network_os = self._network_os or 'default'
        display.display('network_os is set to %s' % self._network_os, log_only=True)

        self._manager = None
        self.key_filename = None

    def exec_command(self, cmd, in_data=None, sudoable=True):
        """Sends the request to the node and returns the reply
        The method accepts two forms of request.  The first form is as a byte
        string that represents xml string be send over netconf session.
        The second form is a json-rpc (2.0) byte string.
        """
        if self._manager:
            # to_ele operates on native strings
            request = to_ele(to_native(cmd, errors='surrogate_or_strict'))

            if request is None:
                return 'unable to parse request'

            try:
                reply = self._manager.rpc(request)
            except RPCError as exc:
                error = self.internal_error(data=to_text(to_xml(exc.xml), errors='surrogate_or_strict'))
                return json.dumps(error)

            return reply.data_xml
        else:
            return super(Connection, self).exec_command(cmd, in_data, sudoable)

    def _connect(self):
        if not HAS_NCCLIENT:
            raise AnsibleError(
                'ncclient is required to use the netconf connection type.\n'
                'Please run pip install ncclient'
            )

        super(Connection, self)._connect()
        display.display('ssh connection done, starting ncclient', log_only=True)

        allow_agent = True
        if self._play_context.password is not None:
            allow_agent = False
        setattr(self._play_context, 'allow_agent', allow_agent)

        self.key_filename = self._play_context.private_key_file or self.get_option('private_key_file')
        if self.key_filename:
            self.key_filename = str(os.path.expanduser(self.key_filename))

        if self._network_os == 'default':
            for cls in netconf_loader.all(class_only=True):
                network_os = cls.guess_network_os(self)
                if network_os:
                    display.display('discovered network_os %s' % network_os, log_only=True)
                    self._network_os = network_os

        device_params = {'name': NETWORK_OS_DEVICE_PARAM_MAP.get(self._network_os) or self._network_os}

        ssh_config = self.get_option('netconf_ssh_config')
        if ssh_config in BOOLEANS_TRUE:
            ssh_config = True
        elif ssh_config in BOOLEANS_FALSE:
            ssh_config = None

        try:
            port = self._play_context.port or 830
            display.vvvv("ESTABLISH NETCONF SSH CONNECTION FOR USER: %s on PORT %s TO %s" %
                         (self._play_context.remote_user, port, self._play_context.remote_addr))
            self._manager = manager.connect(
                host=self._play_context.remote_addr,
                port=port,
                username=self._play_context.remote_user,
                password=self._play_context.password,
                key_filename=self.key_filename,
                hostkey_verify=self.get_option('host_key_checking'),
                look_for_keys=self.get_option('look_for_keys'),
                device_params=device_params,
                allow_agent=self._play_context.allow_agent,
                timeout=self._play_context.timeout,
                ssh_config=ssh_config
            )
        except SSHUnknownHostError as exc:
            raise AnsibleConnectionFailure(str(exc))
        except ImportError as exc:
            raise AnsibleError("connection=netconf is not supported on {0}".format(self._network_os))

        if not self._manager.connected:
            return 1, b'', b'not connected'

        display.display('ncclient manager object created successfully', log_only=True)

        self._connected = True

        netconf = netconf_loader.get(self._network_os, self)
        if netconf:
            display.display('loaded netconf plugin for network_os %s' % self._network_os, log_only=True)
        else:
            netconf = netconf_loader.get("default", self)
            display.display('unable to load netconf plugin for network_os %s, falling back to default plugin' % self._network_os)
        self._implementation_plugins.append(netconf)

        super(Connection, self)._connect()

        return 0, to_bytes(self._manager.session_id, errors='surrogate_or_strict'), b''

    def close(self):
        if self._manager:
            self._manager.close_session()
        super(Connection, self).close()



#!/usr/bin/python
# (c) 2015, Werner Dijkerman (ikben@werner-dijkerman.nl)
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}

DOCUMENTATION = '''
---
module: gitlab_group
short_description: Creates/updates/deletes Gitlab Groups
description:
   - When the group does not exist in Gitlab, it will be created.
   - When the group does exist and state=absent, the group will be deleted.
   - As of Ansible version 2.7, this module make use of a different python module and thus some arguments are deprecated.
version_added: "2.1"
author: "Werner Dijkerman (@dj-wasabi)"
requirements:
    - python-gitlab python module
options:
    server_url:
        description:
            - Url of Gitlab server, with protocol (http or https).
        required: true
    validate_certs:
        description:
            - When using https if SSL certificate needs to be verified.
        required: false
        default: true
        aliases:
            - verify_ssl
    login_user:
        description:
            - Gitlab user name.
        required: false
        default: null
    login_password:
        description:
            - Gitlab password for login_user
        required: false
        default: null
    login_token:
        description:
            - Gitlab token for logging in.
        required: false
        default: null
    name:
        description:
            - Name of the group you want to create.
        required: true
    path:
        description:
            - The path of the group you want to create, this will be server_url/group_path
            - If not supplied, the group_name will be used.
        required: false
        default: null
    description:
        description:
            - A description for the group.
        required: false
        default: null
        version_added: "2.7"
    state:
        description:
            - create or delete group.
            - Possible values are present and absent.
        required: false
        default: "present"
        choices: ["present", "absent"]
'''

EXAMPLES = '''
- name: "Delete Gitlab Group"
  local_action:
    gitlab_group:
        server_url: http://gitlab.dj-wasabi.local
        validate_certs: False
        login_token: WnUzDsxjy8230-Dy_k
        name: my_first_group
        state: absent

- name: "Create Gitlab Group"
  local_action:
    gitlab_group:
        server_url: https://gitlab.dj-wasabi.local"
        validate_certs: True
        login_user: dj-wasabi
        login_password: "MySecretPassword"
        name: my_first_group
        path: my_first_group
        state: present
'''

RETURN = '''# '''

try:
    import gitlab
    HAS_GITLAB_PACKAGE = True
except:
    HAS_GITLAB_PACKAGE = False

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils._text import to_native


class GitLabGroup(object):
    def __init__(self, module, git):
        self._module = module
        self._gitlab = git
        self.groupObject = None

    def createOrUpdateGroup(self, name, path, description):
        changed = False
        if self.groupObject is None:
            group = self._gitlab.groups.create({'name': name, 'path': path})
            changed = True
        else:
            group = self.groupObject

        if description is not None:
            if group.description != description:
                group.description = description
                changed = True

        if changed:
            if self._module.check_mode:
                self._module.exit_json(changed=True, result="Group should have updated.")
            try:
                group.save()
            except Exception as e:
                self._module.fail_json(msg="Failed to create or update a group: %s " % e)
            return True
        else:
            return False

    def deleteGroup(self):
        group = self.groupObject
        if len(group.projects.list()) >= 1:
            self._module.fail_json(
                msg="There are still projects in this group. These needs to be moved or deleted before this group can be removed.")
        else:
            if self._module.check_mode:
                self._module.exit_json(changed=True)
            try:
                group.delete()
            except Exception as e:
                self._module.fail_json(msg="Failed to delete a group: %s " % e)
            return True

    def existsGroup(self, name):
        """When group/user exists, object will be stored in self.groupObject."""
        groups = self._gitlab.groups.list(search=name)
        if len(groups) == 1:
            self.groupObject = groups[0]
            return True


def main():
    module = AnsibleModule(
        argument_spec=dict(
            server_url=dict(required=True, type='str'),
            validate_certs=dict(required=False, default=True, type='bool', aliases=['verify_ssl']),
            login_user=dict(required=False, no_log=True, type='str'),
            login_password=dict(required=False, no_log=True, type='str'),
            login_token=dict(required=False, no_log=True, type='str'),
            name=dict(required=True, type='str'),
            path=dict(required=False, type='str'),
            description=dict(required=False, type='str'),
            state=dict(default="present", choices=["present", "absent"]),
        ),
        mutually_exclusive=[
            ['login_user', 'login_token'],
            ['login_password', 'login_token']
        ],
        required_together=[
            ['login_user', 'login_password']
        ],
        required_one_of=[
            ['login_user', 'login_token']
        ],
        supports_check_mode=True
    )

    if not HAS_GITLAB_PACKAGE:
        module.fail_json(msg="Missing required gitlab module (check docs or install with: pip install python-gitlab")

    server_url = module.params['server_url']
    validate_certs = module.params['validate_certs']
    login_user = module.params['login_user']
    login_password = module.params['login_password']
    login_token = module.params['login_token']
    group_name = module.params['name']
    group_path = module.params['path']
    description = module.params['description']
    state = module.params['state']

    try:
        git = gitlab.Gitlab(url=server_url, ssl_verify=validate_certs, email=login_user, password=login_password,
                            private_token=login_token, api_version=4)
        git.auth()
    except (gitlab.exceptions.GitlabAuthenticationError, gitlab.exceptions.GitlabGetError) as e:
        module.fail_json(msg='Failed to connect to Gitlab server: %s' % to_native(e))

    if group_path is None:
        group_path = group_name.replace(" ", "_")

    group = GitLabGroup(module, git)
    group_name = group_name.lower()
    group_exists = group.existsGroup(group_name)

    if group_exists and state == "absent":
        if group.deleteGroup():
            module.exit_json(changed=True, result="Successfully deleted group %s" % group_name)
    else:
        if state == "absent":
            module.exit_json(changed=False, result="Group deleted or does not exists")
        else:
            if group.createOrUpdateGroup(name=group_name, path=group_path, description=description):
                module.exit_json(changed=True, result="Successfully created or updated the group %s" % group_name)
            else:
                module.exit_json(changed=False, result="No need to update the group %s" % group_name)


if __name__ == '__main__':
    main()



#
# (c) 2016 Red Hat Inc.
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.

# Make coding more python3-ish
from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

from ansible.compat.tests.mock import patch
from ansible.modules.network.ovs import openvswitch_db
from units.modules.utils import set_module_args
from .ovs_module import TestOpenVSwitchModule, load_fixture

test_name_side_effect_matrix = {
    'test_openvswitch_db_absent_idempotent': [
        (0, 'openvswitch_db_disable_in_band_missing.cfg', None),
        (0, None, None)],
    'test_openvswitch_db_absent_removes_key': [
        (0, 'openvswitch_db_disable_in_band_true.cfg', None),
        (0, None, None)],
    'test_openvswitch_db_present_idempotent': [
        (0, 'openvswitch_db_disable_in_band_true.cfg', None),
        (0, None, None)],
    'test_openvswitch_db_present_adds_key': [
        (0, 'openvswitch_db_disable_in_band_missing.cfg', None),
        (0, None, None)],
    'test_openvswitch_db_present_updates_key': [
        (0, 'openvswitch_db_disable_in_band_true.cfg', None),
        (0, None, None)],
    'test_openvswitch_db_present_missing_key_on_map': [
        (0, 'openvswitch_db_disable_in_band_true.cfg', None),
        (0, None, None)],
    'test_openvswitch_db_present_stp_enable': [
        (0, 'openvswitch_db_disable_in_band_true.cfg', None),
        (0, None, None)],
}


class TestOpenVSwitchDBModule(TestOpenVSwitchModule):

    module = openvswitch_db

    def setUp(self):
        super(TestOpenVSwitchDBModule, self).setUp()

        self.mock_run_command = (
            patch('ansible.module_utils.basic.AnsibleModule.run_command'))
        self.run_command = self.mock_run_command.start()
        self.mock_get_bin_path = (
            patch('ansible.module_utils.basic.AnsibleModule.get_bin_path'))
        self.get_bin_path = self.mock_get_bin_path.start()

    def tearDown(self):
        super(TestOpenVSwitchDBModule, self).tearDown()

        self.mock_run_command.stop()
        self.mock_get_bin_path.stop()

    def load_fixtures(self, test_name):
        test_side_effects = []
        for s in test_name_side_effect_matrix[test_name]:
            rc = s[0]
            out = load_fixture(s[1]) if s[1] else None
            err = s[2]
            side_effect_with_fixture_loaded = (rc, out, err)
            test_side_effects.append(side_effect_with_fixture_loaded)
        self.run_command.side_effect = test_side_effects

        self.get_bin_path.return_value = '/usr/bin/ovs-vsctl'

    def test_openvswitch_db_absent_idempotent(self):
        set_module_args(dict(state='absent',
                             table='Bridge', record='test-br',
                             col='other_config', key='disable-in-band',
                             value='True'))
        self.execute_module(test_name='test_openvswitch_db_absent_idempotent')

    def test_openvswitch_db_absent_removes_key(self):
        set_module_args(dict(state='absent',
                             table='Bridge', record='test-br',
                             col='other_config', key='disable-in-band',
                             value='True'))
        self.execute_module(
            changed=True,
            commands=['/usr/bin/ovs-vsctl -t 5 remove Bridge test-br other_config'
                      ' disable-in-band=True'],
            test_name='test_openvswitch_db_absent_removes_key')

    def test_openvswitch_db_present_idempotent(self):
        set_module_args(dict(state='present',
                             table='Bridge', record='test-br',
                             col='other_config', key='disable-in-band',
                             value='True'))
        self.execute_module(test_name='test_openvswitch_db_present_idempotent')

    def test_openvswitch_db_present_adds_key(self):
        set_module_args(dict(state='present',
                             table='Bridge', record='test-br',
                             col='other_config', key='disable-in-band',
                             value='True'))
        self.execute_module(
            changed=True,
            commands=['/usr/bin/ovs-vsctl -t 5 set Bridge test-br other_config'
                      ':disable-in-band=True'],
            test_name='test_openvswitch_db_present_adds_key')

    def test_openvswitch_db_present_updates_key(self):
        set_module_args(dict(state='present',
                             table='Bridge', record='test-br',
                             col='other_config', key='disable-in-band',
                             value='False'))
        self.execute_module(
            changed=True,
            commands=['/usr/bin/ovs-vsctl -t 5 set Bridge test-br other_config'
                      ':disable-in-band=False'],
            test_name='test_openvswitch_db_present_updates_key')

    def test_openvswitch_db_present_missing_key_on_map(self):
        set_module_args(dict(state='present',
                             table='Bridge', record='test-br',
                             col='other_config',
                             value='False'))
        self.execute_module(
            failed=True,
            test_name='test_openvswitch_db_present_idempotent')

    def test_openvswitch_db_present_stp_enable(self):
        set_module_args(dict(state='present',
                             table='Bridge', record='test-br',
                             col='stp_enable',
                             value='False'))
        self.execute_module(changed=True,
                            test_name='test_openvswitch_db_present_stp_enable')



#!/usr/bin/python
# -*- coding: utf-8 -*-

# (c) 2015, Marius Gedminas <marius@pov.lt>
# (c) 2016, Matthew Gamble <git@matthewgamble.net>
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}


DOCUMENTATION = '''
---
module: git_config
author:
  - Matthew Gamble (@djmattyg007)
  - Marius Gedminas
version_added: 2.1
requirements: ['git']
short_description: Read and write git configuration
description:
  - The C(git_config) module changes git configuration by invoking 'git config'.
    This is needed if you don't want to use M(template) for the entire git
    config file (e.g. because you need to change just C(user.email) in
    /etc/.git/config).  Solutions involving M(command) are cumbersome or
    don't work correctly in check mode.
options:
  list_all:
    description:
      - List all settings (optionally limited to a given I(scope))
    type: bool
    default: 'no'
  name:
    description:
      - The name of the setting. If no value is supplied, the value will
        be read from the config if it has been set.
  repo:
    description:
      - Path to a git repository for reading and writing values from a
        specific repo.
  scope:
    description:
      - Specify which scope to read/set values from. This is required
        when setting config values. If this is set to local, you must
        also specify the repo parameter. It defaults to system only when
        not using I(list_all)=yes.
    choices: [ "local", "global", "system" ]
  value:
    description:
      - When specifying the name of a single setting, supply a value to
        set that setting to the given value.
'''

EXAMPLES = '''
# Set some settings in ~/.gitconfig
- git_config:
    name: alias.ci
    scope: global
    value: commit

- git_config:
    name: alias.st
    scope: global
    value: status

# Or system-wide:
- git_config:
    name: alias.remotev
    scope: system
    value: remote -v

- git_config:
    name: core.editor
    scope: global
    value: vim

# scope=system is the default
- git_config:
    name: alias.diffc
    value: diff --cached

- git_config:
    name: color.ui
    value: auto

# Make etckeeper not complain when invoked by cron
- git_config:
    name: user.email
    repo: /etc
    scope: local
    value: 'root@{{ ansible_fqdn }}'

# Read individual values from git config
- git_config:
    name: alias.ci
    scope: global

# scope: system is also assumed when reading values, unless list_all=yes
- git_config:
    name: alias.diffc

# Read all values from git config
- git_config:
    list_all: yes
    scope: global

# When list_all=yes and no scope is specified, you get configuration from all scopes
- git_config:
    list_all: yes

# Specify a repository to include local settings
- git_config:
    list_all: yes
    repo: /path/to/repo.git
'''

RETURN = '''
---
config_value:
  description: When list_all=no and value is not set, a string containing the value of the setting in name
  returned: success
  type: string
  sample: "vim"

config_values:
  description: When list_all=yes, a dict containing key/value pairs of multiple configuration settings
  returned: success
  type: dictionary
  sample:
    core.editor: "vim"
    color.ui: "auto"
    alias.diffc: "diff --cached"
    alias.remotev: "remote -v"
'''
from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.six.moves import shlex_quote


def main():
    module = AnsibleModule(
        argument_spec=dict(
            list_all=dict(required=False, type='bool', default=False),
            name=dict(type='str'),
            repo=dict(type='path'),
            scope=dict(required=False, type='str', choices=['local', 'global', 'system']),
            value=dict(required=False)
        ),
        mutually_exclusive=[['list_all', 'name'], ['list_all', 'value']],
        required_if=[('scope', 'local', ['repo'])],
        required_one_of=[['list_all', 'name']],
        supports_check_mode=True,
    )
    git_path = module.get_bin_path('git', True)

    params = module.params
    # We check error message for a pattern, so we need to make sure the messages appear in the form we're expecting.
    # Set the locale to C to ensure consistent messages.
    module.run_command_environ_update = dict(LANG='C', LC_ALL='C', LC_MESSAGES='C', LC_CTYPE='C')

    if params['name']:
        name = params['name']
    else:
        name = None

    if params['scope']:
        scope = params['scope']
    elif params['list_all']:
        scope = None
    else:
        scope = 'system'

    if params['value']:
        new_value = params['value']
    else:
        new_value = None

    args = [git_path, "config", "--includes"]
    if params['list_all']:
        args.append('-l')
    if scope:
        args.append("--" + scope)
    if name:
        args.append(name)

    if scope == 'local':
        dir = params['repo']
    elif params['list_all'] and params['repo']:
        # Include local settings from a specific repo when listing all available settings
        dir = params['repo']
    else:
        # Run from root directory to avoid accidentally picking up any local config settings
        dir = "/"

    (rc, out, err) = module.run_command(' '.join(args), cwd=dir)
    if params['list_all'] and scope and rc == 128 and 'unable to read config file' in err:
        # This just means nothing has been set at the given scope
        module.exit_json(changed=False, msg='', config_values={})
    elif rc >= 2:
        # If the return code is 1, it just means the option hasn't been set yet, which is fine.
        module.fail_json(rc=rc, msg=err, cmd=' '.join(args))

    if params['list_all']:
        values = out.rstrip().splitlines()
        config_values = {}
        for value in values:
            k, v = value.split('=', 1)
            config_values[k] = v
        module.exit_json(changed=False, msg='', config_values=config_values)
    elif not new_value:
        module.exit_json(changed=False, msg='', config_value=out.rstrip())
    else:
        old_value = out.rstrip()
        if old_value == new_value:
            module.exit_json(changed=False, msg="")

    if not module.check_mode:
        new_value_quoted = shlex_quote(new_value)
        cmd = ' '.join(args + [new_value_quoted])
        (rc, out, err) = module.run_command(cmd, cwd=dir)
        if err:
            module.fail_json(rc=rc, msg=err, cmd=cmd)

    module.exit_json(
        msg='setting changed',
        diff=dict(
            before_header=' '.join(args),
            before=old_value + "\n",
            after_header=' '.join(args),
            after=new_value + "\n"
        ),
        changed=True
    )


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# (c) 2013, Chatham Financial <oss@chathamfinancial.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}


DOCUMENTATION = '''
---
module: rabbitmq_user
short_description: Adds or removes users to RabbitMQ
description:
  - Add or remove users to RabbitMQ and assign permissions
version_added: "1.1"
author: '"Chris Hoffman (@chrishoffman)"'
options:
  user:
    description:
      - Name of user to add
    required: true
    aliases: [username, name]
  password:
    description:
      - Password of user to add.
      - To change the password of an existing user, you must also specify
        C(update_password=always).
  tags:
    description:
      - User tags specified as comma delimited
  permissions:
    description:
      - a list of dicts, each dict contains vhost, configure_priv, write_priv, and read_priv,
        and represents a permission rule for that vhost.
      - This option should be preferable when you care about all permissions of the user.
      - You should use vhost, configure_priv, write_priv, and read_priv options instead
        if you care about permissions for just some vhosts.
    default: []
  vhost:
    description:
      - vhost to apply access privileges.
      - This option will be ignored when permissions option is used.
    default: /
  node:
    description:
      - erlang node name of the rabbit we wish to configure
    default: rabbit
    version_added: "1.2"
  configure_priv:
    description:
      - Regular expression to restrict configure actions on a resource
        for the specified vhost.
      - By default all actions are restricted.
      - This option will be ignored when permissions option is used.
    default: ^$
  write_priv:
    description:
      - Regular expression to restrict configure actions on a resource
        for the specified vhost.
      - By default all actions are restricted.
      - This option will be ignored when permissions option is used.
    default: ^$
  read_priv:
    description:
      - Regular expression to restrict configure actions on a resource
        for the specified vhost.
      - By default all actions are restricted.
      - This option will be ignored when permissions option is used.
    default: ^$
  force:
    description:
      - Deletes and recreates the user.
    type: bool
    default: 'no'
  state:
    description:
      - Specify if user is to be added or removed
    default: present
    choices: [present, absent]
  update_password:
    description:
      - C(on_create) will only set the password for newly created users.  C(always) will update passwords if they differ.
    required: false
    default: on_create
    choices: [ on_create, always ]
    version_added: "2.6"
'''

EXAMPLES = '''
# Add user to server and assign full access control on / vhost.
# The user might have permission rules for other vhost but you don't care.
- rabbitmq_user:
    user: joe
    password: changeme
    vhost: /
    configure_priv: .*
    read_priv: .*
    write_priv: .*
    state: present

# Add user to server and assign full access control on / vhost.
# The user doesn't have permission rules for other vhosts
- rabbitmq_user:
    user: joe
    password: changeme
    permissions:
      - vhost: /
        configure_priv: .*
        read_priv: .*
        write_priv: .*
    state: present
'''

from ansible.module_utils.basic import AnsibleModule


class RabbitMqUser(object):
    def __init__(self, module, username, password, tags, permissions,
                 node, bulk_permissions=False):
        self.module = module
        self.username = username
        self.password = password
        self.node = node
        if not tags:
            self.tags = list()
        else:
            self.tags = tags.split(',')

        self.permissions = permissions
        self.bulk_permissions = bulk_permissions

        self._tags = None
        self._permissions = []
        self._rabbitmqctl = module.get_bin_path('rabbitmqctl', True)

    def _exec(self, args, run_in_check_mode=False):
        if not self.module.check_mode or run_in_check_mode:
            cmd = [self._rabbitmqctl, '-q']
            if self.node is not None:
                cmd.extend(['-n', self.node])
            rc, out, err = self.module.run_command(cmd + args, check_rc=True)
            return out.splitlines()
        return list()

    def get(self):
        users = self._exec(['list_users'], True)

        for user_tag in users:
            if '\t' not in user_tag:
                continue

            user, tags = user_tag.split('\t')

            if user == self.username:
                for c in ['[', ']', ' ']:
                    tags = tags.replace(c, '')

                if tags != '':
                    self._tags = tags.split(',')
                else:
                    self._tags = list()

                self._permissions = self._get_permissions()
                return True
        return False

    def _get_permissions(self):
        perms_out = [perm for perm in self._exec(['list_user_permissions', self.username], True) if perm.strip()]

        perms_list = list()
        for perm in perms_out:
            vhost, configure_priv, write_priv, read_priv = perm.split('\t')
            if not self.bulk_permissions:
                if vhost == self.permissions[0]['vhost']:
                    perms_list.append(dict(vhost=vhost, configure_priv=configure_priv,
                                           write_priv=write_priv, read_priv=read_priv))
                    break
            else:
                perms_list.append(dict(vhost=vhost, configure_priv=configure_priv,
                                       write_priv=write_priv, read_priv=read_priv))
        return perms_list

    def check_password(self):
        return self._exec(['authenticate_user', self.username, self.password], True)

    def add(self):
        if self.password is not None:
            self._exec(['add_user', self.username, self.password])
        else:
            self._exec(['add_user', self.username, ''])
            self._exec(['clear_password', self.username])

    def delete(self):
        self._exec(['delete_user', self.username])

    def change_password(self):
        if self.password is not None:
            self._exec(['change_password', self.username, self.password])
        else:
            self._exec(['clear_password', self.username])

    def set_tags(self):
        self._exec(['set_user_tags', self.username] + self.tags)

    def set_permissions(self):
        for permission in self._permissions:
            if permission not in self.permissions:
                cmd = ['clear_permissions', '-p']
                cmd.append(permission['vhost'])
                cmd.append(self.username)
                self._exec(cmd)
        for permission in self.permissions:
            if permission not in self._permissions:
                cmd = ['set_permissions', '-p']
                cmd.append(permission['vhost'])
                cmd.append(self.username)
                cmd.append(permission['configure_priv'])
                cmd.append(permission['write_priv'])
                cmd.append(permission['read_priv'])
                self._exec(cmd)

    def has_tags_modifications(self):
        return set(self.tags) != set(self._tags)

    def has_permissions_modifications(self):
        return sorted(self._permissions) != sorted(self.permissions)


def main():
    arg_spec = dict(
        user=dict(required=True, aliases=['username', 'name']),
        password=dict(default=None, no_log=True),
        tags=dict(default=None),
        permissions=dict(default=list(), type='list'),
        vhost=dict(default='/'),
        configure_priv=dict(default='^$'),
        write_priv=dict(default='^$'),
        read_priv=dict(default='^$'),
        force=dict(default='no', type='bool'),
        state=dict(default='present', choices=['present', 'absent']),
        node=dict(default=None),
        update_password=dict(default='on_create', choices=['on_create', 'always'])
    )
    module = AnsibleModule(
        argument_spec=arg_spec,
        supports_check_mode=True
    )

    username = module.params['user']
    password = module.params['password']
    tags = module.params['tags']
    permissions = module.params['permissions']
    vhost = module.params['vhost']
    configure_priv = module.params['configure_priv']
    write_priv = module.params['write_priv']
    read_priv = module.params['read_priv']
    force = module.params['force']
    state = module.params['state']
    node = module.params['node']
    update_password = module.params['update_password']

    bulk_permissions = True
    if not permissions:
        perm = {
            'vhost': vhost,
            'configure_priv': configure_priv,
            'write_priv': write_priv,
            'read_priv': read_priv
        }
        permissions.append(perm)
        bulk_permissions = False

    rabbitmq_user = RabbitMqUser(module, username, password, tags, permissions,
                                 node, bulk_permissions=bulk_permissions)

    result = dict(changed=False, user=username, state=state)

    if rabbitmq_user.get():
        if state == 'absent':
            rabbitmq_user.delete()
            result['changed'] = True
        else:
            if force:
                rabbitmq_user.delete()
                rabbitmq_user.add()
                rabbitmq_user.get()
                result['changed'] = True
            elif update_password == 'always':
                if not rabbitmq_user.check_password():
                    rabbitmq_user.change_password()
                    result['changed'] = True

            if rabbitmq_user.has_tags_modifications():
                rabbitmq_user.set_tags()
                result['changed'] = True

            if rabbitmq_user.has_permissions_modifications():
                rabbitmq_user.set_permissions()
                result['changed'] = True
    elif state == 'present':
        rabbitmq_user.add()
        rabbitmq_user.set_tags()
        rabbitmq_user.set_permissions()
        result['changed'] = True

    module.exit_json(**result)


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017 Google
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# ----------------------------------------------------------------------------
#
#     ***     AUTO GENERATED CODE    ***    AUTO GENERATED CODE     ***
#
# ----------------------------------------------------------------------------
#
#     This file is automatically generated by Magic Modules and manual
#     changes will be clobbered when the file is regenerated.
#
#     Please read more about how to change this file at
#     https://www.github.com/GoogleCloudPlatform/magic-modules
#
# ----------------------------------------------------------------------------

from __future__ import absolute_import, division, print_function
__metaclass__ = type

################################################################################
# Documentation
################################################################################

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ["preview"],
                    'supported_by': 'community'}

DOCUMENTATION = '''
---
module: gcp_sql_instance
description:
    - Represents a Cloud SQL instance. Cloud SQL instances are SQL databases hosted in
      Google's cloud. The Instances resource provides methods for common configuration
      and management tasks.
short_description: Creates a GCP Instance
version_added: 2.7
author: Google Inc. (@googlecloudplatform)
requirements:
    - python >= 2.6
    - requests >= 2.18.4
    - google-auth >= 1.3.0
options:
    state:
        description:
            - Whether the given object should exist in GCP
        choices: ['present', 'absent']
        default: 'present'
    backend_type:
        description:
            - "* FIRST_GEN: First Generation instance. MySQL only."
            - "* SECOND_GEN: Second Generation instance or PostgreSQL instance."
            - "* EXTERNAL: A database server that is not managed by Google."
        required: false
        choices: ['FIRST_GEN', 'SECOND_GEN', 'EXTERNAL']
    connection_name:
        description:
            - Connection name of the Cloud SQL instance used in connection strings.
        required: false
    database_version:
        description:
            - The database engine type and version. For First Generation instances, can be MYSQL_5_5,
              or MYSQL_5_6. For Second Generation instances, can be MYSQL_5_6 or MYSQL_5_7. Defaults
              to MYSQL_5_6.
            - 'PostgreSQL instances: POSTGRES_9_6  The databaseVersion property can not be changed
              after instance creation.'
        required: false
        choices: ['MYSQL_5_5', 'MYSQL_5_6', 'MYSQL_5_7', 'POSTGRES_9_6']
    failover_replica:
        description:
            - The name and status of the failover replica. This property is applicable only to
              Second Generation instances.
        required: false
        suboptions:
            available:
                description:
                    - The availability status of the failover replica. A false status indicates that the
                      failover replica is out of sync. The master can only failover to the falover replica
                      when the status is true.
                required: false
                type: bool
            name:
                description:
                    - The name of the failover replica. If specified at instance creation, a failover
                      replica is created for the instance. The name doesn't include the project ID. This
                      property is applicable only to Second Generation instances.
                required: false
    instance_type:
        description:
            - The instance type. This can be one of the following.
            - "* CLOUD_SQL_INSTANCE: A Cloud SQL instance that is not replicating   from a master."
            - "* ON_PREMISES_INSTANCE: An instance running on the customer's   premises."
            - "* READ_REPLICA_INSTANCE: A Cloud SQL instance configured as a   read-replica."
        required: false
        choices: ['CLOUD_SQL_INSTANCE', 'ON_PREMISES_INSTANCE', 'READ_REPLICA_INSTANCE']
    ipv6_address:
        description:
            - The IPv6 address assigned to the instance. This property is applicable only to First
              Generation instances.
        required: false
    master_instance_name:
        description:
            - The name of the instance which will act as master in the replication setup.
        required: false
    max_disk_size:
        description:
            - The maximum disk size of the instance in bytes.
        required: false
    name:
        description:
            - Name of the Cloud SQL instance. This does not include the project ID.
        required: true
    region:
        description:
            - The geographical region. Defaults to us-central or us-central1 depending on the
              instance type (First Generation or Second Generation/PostgreSQL).
        required: false
    replica_configuration:
        description:
            - Configuration specific to failover replicas and read replicas.
        required: false
        suboptions:
            failover_target:
                description:
                    - Specifies if the replica is the failover target. If the field is set to true the
                      replica will be designated as a failover replica.
                    - In case the master instance fails, the replica instance will be promoted as the
                      new master instance.
                    - Only one replica can be specified as failover target, and the replica has to be
                      in different zone with the master instance.
                required: false
                type: bool
            mysql_replica_configuration:
                description:
                    - MySQL specific configuration when replicating from a MySQL on-premises master. Replication
                      configuration information such as the username, password, certificates, and keys
                      are not stored in the instance metadata.  The configuration information is used
                      only to set up the replication connection and is stored by MySQL in a file named
                      master.info in the data directory.
                required: false
                suboptions:
                    ca_certificate:
                        description:
                            - PEM representation of the trusted CA's x509 certificate.
                        required: false
                    client_certificate:
                        description:
                            - PEM representation of the slave's x509 certificate .
                        required: false
                    client_key:
                        description:
                            - PEM representation of the slave's private key. The corresponsing public key is encoded
                              in the client's asf asd certificate.
                        required: false
                    connect_retry_interval:
                        description:
                            - Seconds to wait between connect retries. MySQL's default is 60 seconds.
                        required: false
                    dump_file_path:
                        description:
                            - Path to a SQL dump file in Google Cloud Storage from which the slave instance is
                              to be created. The URI is in the form gs://bucketName/fileName. Compressed gzip
                              files (.gz) are also supported. Dumps should have the binlog co-ordinates from which
                              replication should begin. This can be accomplished by setting --master-data to 1
                              when using mysqldump.
                        required: false
                    master_heartbeat_period:
                        description:
                            - Interval in milliseconds between replication heartbeats.
                        required: false
                    password:
                        description:
                            - The password for the replication connection.
                        required: false
                    ssl_cipher:
                        description:
                            - A list of permissible ciphers to use for SSL encryption.
                        required: false
                    username:
                        description:
                            - The username for the replication connection.
                        required: false
                    verify_server_certificate:
                        description:
                            - Whether or not to check the master's Common Name value in the certificate that it
                              sends during the SSL handshake.
                        required: false
                        type: bool
            replica_names:
                description:
                    - The replicas of the instance.
                required: false
            service_account_email_address:
                description:
                    - The service account email address assigned to the instance. This property is applicable
                      only to Second Generation instances.
                required: false
    settings:
        description:
            - The user settings.
        required: false
        suboptions:
            ip_configuration:
                description:
                    - The settings for IP Management. This allows to enable or disable the instance IP
                      and manage which external networks can connect to the instance. The IPv4 address
                      cannot be disabled for Second Generation instances.
                required: false
                suboptions:
                    ipv4_enabled:
                        description:
                            - Whether the instance should be assigned an IP address or not.
                        required: false
                        type: bool
                    authorized_networks:
                        description:
                            - The list of external networks that are allowed to connect to the instance using
                              the IP. In CIDR notation, also known as 'slash' notation (e.g. 192.168.100.0/24).
                        required: false
                        suboptions:
                            expiration_time:
                                description:
                                    - The time when this access control entry expires in RFC 3339 format, for example
                                      2012-11-15T16:19:00.094Z.
                                required: false
                            name:
                                description:
                                    - An optional label to identify this entry.
                                required: false
                            value:
                                description:
                                    - The whitelisted value for the access control list. For example, to grant access
                                      to a client from an external IP (IPv4 or IPv6) address or subnet, use that address
                                      or subnet here.
                                required: false
                    require_ssl:
                        description:
                            - Whether the mysqld should default to 'REQUIRE X509' for users connecting over IP.
                        required: false
                        type: bool
            tier:
                description:
                    - The tier or machine type for this instance, for example db-n1-standard-1. For MySQL
                      instances, this field determines whether the instance is Second Generation (recommended)
                      or First Generation.
                required: false
extends_documentation_fragment: gcp
'''

EXAMPLES = '''
- name: create a instance
  gcp_sql_instance:
      name: "test_object"
      settings:
        ip_configuration:
          authorized_networks:
          - name: google dns server
            value: 8.8.8.8/32
        tier: db-n1-standard-1
      region: us-central1
      project: "test_project"
      auth_kind: "service_account"
      service_account_file: "/tmp/auth.pem"
      state: present
'''

RETURN = '''
    backend_type:
        description:
            - "* FIRST_GEN: First Generation instance. MySQL only."
            - "* SECOND_GEN: Second Generation instance or PostgreSQL instance."
            - "* EXTERNAL: A database server that is not managed by Google."
        returned: success
        type: str
    connection_name:
        description:
            - Connection name of the Cloud SQL instance used in connection strings.
        returned: success
        type: str
    database_version:
        description:
            - The database engine type and version. For First Generation instances, can be MYSQL_5_5,
              or MYSQL_5_6. For Second Generation instances, can be MYSQL_5_6 or MYSQL_5_7. Defaults
              to MYSQL_5_6.
            - 'PostgreSQL instances: POSTGRES_9_6  The databaseVersion property can not be changed
              after instance creation.'
        returned: success
        type: str
    failover_replica:
        description:
            - The name and status of the failover replica. This property is applicable only to
              Second Generation instances.
        returned: success
        type: complex
        contains:
            available:
                description:
                    - The availability status of the failover replica. A false status indicates that the
                      failover replica is out of sync. The master can only failover to the falover replica
                      when the status is true.
                returned: success
                type: bool
            name:
                description:
                    - The name of the failover replica. If specified at instance creation, a failover
                      replica is created for the instance. The name doesn't include the project ID. This
                      property is applicable only to Second Generation instances.
                returned: success
                type: str
    instance_type:
        description:
            - The instance type. This can be one of the following.
            - "* CLOUD_SQL_INSTANCE: A Cloud SQL instance that is not replicating   from a master."
            - "* ON_PREMISES_INSTANCE: An instance running on the customer's   premises."
            - "* READ_REPLICA_INSTANCE: A Cloud SQL instance configured as a   read-replica."
        returned: success
        type: str
    ip_addresses:
        description:
            - The assigned IP addresses for the instance.
        returned: success
        type: complex
        contains:
            ip_address:
                description:
                    - The IP address assigned.
                returned: success
                type: str
            time_to_retire:
                description:
                    - The due time for this IP to be retired in RFC 3339 format, for example 2012-11-15T16:19:00.094Z.
                      This field is only available when the IP is scheduled to be retired.
                returned: success
                type: str
            type:
                description:
                    - The type of this IP address. A PRIMARY address is an address that can accept incoming
                      connections. An OUTGOING address is the source address of connections originating
                      from the instance, if supported.
                returned: success
                type: str
    ipv6_address:
        description:
            - The IPv6 address assigned to the instance. This property is applicable only to First
              Generation instances.
        returned: success
        type: str
    master_instance_name:
        description:
            - The name of the instance which will act as master in the replication setup.
        returned: success
        type: str
    max_disk_size:
        description:
            - The maximum disk size of the instance in bytes.
        returned: success
        type: int
    name:
        description:
            - Name of the Cloud SQL instance. This does not include the project ID.
        returned: success
        type: str
    region:
        description:
            - The geographical region. Defaults to us-central or us-central1 depending on the
              instance type (First Generation or Second Generation/PostgreSQL).
        returned: success
        type: str
    replica_configuration:
        description:
            - Configuration specific to failover replicas and read replicas.
        returned: success
        type: complex
        contains:
            failover_target:
                description:
                    - Specifies if the replica is the failover target. If the field is set to true the
                      replica will be designated as a failover replica.
                    - In case the master instance fails, the replica instance will be promoted as the
                      new master instance.
                    - Only one replica can be specified as failover target, and the replica has to be
                      in different zone with the master instance.
                returned: success
                type: bool
            mysql_replica_configuration:
                description:
                    - MySQL specific configuration when replicating from a MySQL on-premises master. Replication
                      configuration information such as the username, password, certificates, and keys
                      are not stored in the instance metadata.  The configuration information is used
                      only to set up the replication connection and is stored by MySQL in a file named
                      master.info in the data directory.
                returned: success
                type: complex
                contains:
                    ca_certificate:
                        description:
                            - PEM representation of the trusted CA's x509 certificate.
                        returned: success
                        type: str
                    client_certificate:
                        description:
                            - PEM representation of the slave's x509 certificate .
                        returned: success
                        type: str
                    client_key:
                        description:
                            - PEM representation of the slave's private key. The corresponsing public key is encoded
                              in the client's asf asd certificate.
                        returned: success
                        type: str
                    connect_retry_interval:
                        description:
                            - Seconds to wait between connect retries. MySQL's default is 60 seconds.
                        returned: success
                        type: int
                    dump_file_path:
                        description:
                            - Path to a SQL dump file in Google Cloud Storage from which the slave instance is
                              to be created. The URI is in the form gs://bucketName/fileName. Compressed gzip
                              files (.gz) are also supported. Dumps should have the binlog co-ordinates from which
                              replication should begin. This can be accomplished by setting --master-data to 1
                              when using mysqldump.
                        returned: success
                        type: str
                    master_heartbeat_period:
                        description:
                            - Interval in milliseconds between replication heartbeats.
                        returned: success
                        type: int
                    password:
                        description:
                            - The password for the replication connection.
                        returned: success
                        type: str
                    ssl_cipher:
                        description:
                            - A list of permissible ciphers to use for SSL encryption.
                        returned: success
                        type: str
                    username:
                        description:
                            - The username for the replication connection.
                        returned: success
                        type: str
                    verify_server_certificate:
                        description:
                            - Whether or not to check the master's Common Name value in the certificate that it
                              sends during the SSL handshake.
                        returned: success
                        type: bool
            replica_names:
                description:
                    - The replicas of the instance.
                returned: success
                type: list
            service_account_email_address:
                description:
                    - The service account email address assigned to the instance. This property is applicable
                      only to Second Generation instances.
                returned: success
                type: str
    settings:
        description:
            - The user settings.
        returned: success
        type: complex
        contains:
            ip_configuration:
                description:
                    - The settings for IP Management. This allows to enable or disable the instance IP
                      and manage which external networks can connect to the instance. The IPv4 address
                      cannot be disabled for Second Generation instances.
                returned: success
                type: complex
                contains:
                    ipv4_enabled:
                        description:
                            - Whether the instance should be assigned an IP address or not.
                        returned: success
                        type: bool
                    authorized_networks:
                        description:
                            - The list of external networks that are allowed to connect to the instance using
                              the IP. In CIDR notation, also known as 'slash' notation (e.g. 192.168.100.0/24).
                        returned: success
                        type: complex
                        contains:
                            expiration_time:
                                description:
                                    - The time when this access control entry expires in RFC 3339 format, for example
                                      2012-11-15T16:19:00.094Z.
                                returned: success
                                type: str
                            name:
                                description:
                                    - An optional label to identify this entry.
                                returned: success
                                type: str
                            value:
                                description:
                                    - The whitelisted value for the access control list. For example, to grant access
                                      to a client from an external IP (IPv4 or IPv6) address or subnet, use that address
                                      or subnet here.
                                returned: success
                                type: str
                    require_ssl:
                        description:
                            - Whether the mysqld should default to 'REQUIRE X509' for users connecting over IP.
                        returned: success
                        type: bool
            tier:
                description:
                    - The tier or machine type for this instance, for example db-n1-standard-1. For MySQL
                      instances, this field determines whether the instance is Second Generation (recommended)
                      or First Generation.
                returned: success
                type: str
'''

################################################################################
# Imports
################################################################################

from ansible.module_utils.gcp_utils import navigate_hash, GcpSession, GcpModule, GcpRequest, remove_nones_from_dict, replace_resource_dict
import json
import time

################################################################################
# Main
################################################################################


def main():
    """Main function"""

    module = GcpModule(
        argument_spec=dict(
            state=dict(default='present', choices=['present', 'absent'], type='str'),
            backend_type=dict(type='str', choices=['FIRST_GEN', 'SECOND_GEN', 'EXTERNAL']),
            connection_name=dict(type='str'),
            database_version=dict(type='str', choices=['MYSQL_5_5', 'MYSQL_5_6', 'MYSQL_5_7', 'POSTGRES_9_6']),
            failover_replica=dict(type='dict', options=dict(
                available=dict(type='bool'),
                name=dict(type='str')
            )),
            instance_type=dict(type='str', choices=['CLOUD_SQL_INSTANCE', 'ON_PREMISES_INSTANCE', 'READ_REPLICA_INSTANCE']),
            ipv6_address=dict(type='str'),
            master_instance_name=dict(type='str'),
            max_disk_size=dict(type='int'),
            name=dict(required=True, type='str'),
            region=dict(type='str'),
            replica_configuration=dict(type='dict', options=dict(
                failover_target=dict(type='bool'),
                mysql_replica_configuration=dict(type='dict', options=dict(
                    ca_certificate=dict(type='str'),
                    client_certificate=dict(type='str'),
                    client_key=dict(type='str'),
                    connect_retry_interval=dict(type='int'),
                    dump_file_path=dict(type='str'),
                    master_heartbeat_period=dict(type='int'),
                    password=dict(type='str'),
                    ssl_cipher=dict(type='str'),
                    username=dict(type='str'),
                    verify_server_certificate=dict(type='bool')
                )),
                replica_names=dict(type='list', elements='str'),
                service_account_email_address=dict(type='str')
            )),
            settings=dict(type='dict', options=dict(
                ip_configuration=dict(type='dict', options=dict(
                    ipv4_enabled=dict(type='bool'),
                    authorized_networks=dict(type='list', elements='dict', options=dict(
                        expiration_time=dict(type='str'),
                        name=dict(type='str'),
                        value=dict(type='str')
                    )),
                    require_ssl=dict(type='bool')
                )),
                tier=dict(type='str')
            ))
        )
    )

    if not module.params['scopes']:
        module.params['scopes'] = ['https://www.googleapis.com/auth/sqlservice.admin']

    state = module.params['state']
    kind = 'sql#instance'

    fetch = fetch_resource(module, self_link(module), kind)
    changed = False

    if fetch:
        if state == 'present':
            if is_different(module, fetch):
                fetch = update(module, self_link(module), kind, fetch)
                changed = True
        else:
            delete(module, self_link(module), kind, fetch)
            fetch = {}
            changed = True
    else:
        if state == 'present':
            fetch = create(module, collection(module), kind)
            changed = True
        else:
            fetch = {}

    fetch.update({'changed': changed})

    module.exit_json(**fetch)


def create(module, link, kind):
    auth = GcpSession(module, 'sql')
    return wait_for_operation(module, auth.post(link, resource_to_request(module)))


def update(module, link, kind, fetch):
    auth = GcpSession(module, 'sql')
    return wait_for_operation(module, auth.put(link, resource_to_request(module)))


def delete(module, link, kind, fetch):
    auth = GcpSession(module, 'sql')
    return wait_for_operation(module, auth.delete(link))


def resource_to_request(module):
    request = {
        u'kind': 'sql#instance',
        u'backendType': module.params.get('backend_type'),
        u'connectionName': module.params.get('connection_name'),
        u'databaseVersion': module.params.get('database_version'),
        u'failoverReplica': InstanceFailoverReplica(module.params.get('failover_replica', {}), module).to_request(),
        u'instanceType': module.params.get('instance_type'),
        u'ipv6Address': module.params.get('ipv6_address'),
        u'masterInstanceName': module.params.get('master_instance_name'),
        u'maxDiskSize': module.params.get('max_disk_size'),
        u'name': module.params.get('name'),
        u'region': module.params.get('region'),
        u'replicaConfiguration': InstanceReplicaConfiguration(module.params.get('replica_configuration', {}), module).to_request(),
        u'settings': InstanceSettings(module.params.get('settings', {}), module).to_request()
    }
    return_vals = {}
    for k, v in request.items():
        if v:
            return_vals[k] = v

    return return_vals


def fetch_resource(module, link, kind):
    auth = GcpSession(module, 'sql')
    return return_if_object(module, auth.get(link), kind)


def self_link(module):
    return "https://www.googleapis.com/sql/v1beta4/projects/{project}/instances/{name}".format(**module.params)


def collection(module):
    return "https://www.googleapis.com/sql/v1beta4/projects/{project}/instances".format(**module.params)


def return_if_object(module, response, kind):
    # If not found, return nothing.
    if response.status_code == 404:
        return None

    # If no content, return nothing.
    if response.status_code == 204:
        return None

    # SQL only: return on 403 if not exist
    if response.status_code == 403:
        return None

    try:
        result = response.json()
    except getattr(json.decoder, 'JSONDecodeError', ValueError) as inst:
        module.fail_json(msg="Invalid JSON response with error: %s" % inst)

    if navigate_hash(result, ['error', 'errors']):
        module.fail_json(msg=navigate_hash(result, ['error', 'errors']))
    if result['kind'] != kind:
        module.fail_json(msg="Incorrect result: {kind}".format(**result))

    return result


def is_different(module, response):
    request = resource_to_request(module)
    response = response_to_hash(module, response)

    # Remove all output-only from response.
    response_vals = {}
    for k, v in response.items():
        if k in request:
            response_vals[k] = v

    request_vals = {}
    for k, v in request.items():
        if k in response:
            request_vals[k] = v

    return GcpRequest(request_vals) != GcpRequest(response_vals)


# Remove unnecessary properties from the response.
# This is for doing comparisons with Ansible's current parameters.
def response_to_hash(module, response):
    return {
        u'backendType': response.get(u'backendType'),
        u'connectionName': response.get(u'connectionName'),
        u'databaseVersion': response.get(u'databaseVersion'),
        u'failoverReplica': InstanceFailoverReplica(response.get(u'failoverReplica', {}), module).from_response(),
        u'instanceType': response.get(u'instanceType'),
        u'ipAddresses': InstanceIpAddressesArray(response.get(u'ipAddresses', []), module).from_response(),
        u'ipv6Address': response.get(u'ipv6Address'),
        u'masterInstanceName': response.get(u'masterInstanceName'),
        u'maxDiskSize': response.get(u'maxDiskSize'),
        u'name': response.get(u'name'),
        u'region': response.get(u'region'),
        u'replicaConfiguration': InstanceReplicaConfiguration(response.get(u'replicaConfiguration', {}), module).from_response(),
        u'settings': InstanceSettings(response.get(u'settings', {}), module).from_response()
    }


def async_op_url(module, extra_data=None):
    if extra_data is None:
        extra_data = {}
    url = "https://www.googleapis.com/sql/v1beta4/projects/{project}/operations/{op_id}"
    combined = extra_data.copy()
    combined.update(module.params)
    return url.format(**combined)


def wait_for_operation(module, response):
    op_result = return_if_object(module, response, 'sql#operation')
    if op_result is None:
        return {}
    status = navigate_hash(op_result, ['status'])
    wait_done = wait_for_completion(status, op_result, module)
    return fetch_resource(module, navigate_hash(wait_done, ['targetLink']), 'sql#instance')


def wait_for_completion(status, op_result, module):
    op_id = navigate_hash(op_result, ['name'])
    op_uri = async_op_url(module, {'op_id': op_id})
    while status != 'DONE':
        raise_if_errors(op_result, ['error', 'errors'], 'message')
        time.sleep(1.0)
        if status not in ['PENDING', 'RUNNING', 'DONE']:
            module.fail_json(msg="Invalid result %s" % status)
        op_result = fetch_resource(module, op_uri, 'sql#operation')
        status = navigate_hash(op_result, ['status'])
    return op_result


def raise_if_errors(response, err_path, module):
    errors = navigate_hash(response, err_path)
    if errors is not None:
        module.fail_json(msg=errors)


class InstanceFailoverReplica(object):
    def __init__(self, request, module):
        self.module = module
        if request:
            self.request = request
        else:
            self.request = {}

    def to_request(self):
        return remove_nones_from_dict({
            u'available': self.request.get('available'),
            u'name': self.request.get('name')
        })

    def from_response(self):
        return remove_nones_from_dict({
            u'available': self.request.get(u'available'),
            u'name': self.request.get(u'name')
        })


class InstanceIpAddressesArray(object):
    def __init__(self, request, module):
        self.module = module
        if request:
            self.request = request
        else:
            self.request = []

    def to_request(self):
        items = []
        for item in self.request:
            items.append(self._request_for_item(item))
        return items

    def from_response(self):
        items = []
        for item in self.request:
            items.append(self._response_from_item(item))
        return items

    def _request_for_item(self, item):
        return remove_nones_from_dict({
            u'ipAddress': item.get('ip_address'),
            u'timeToRetire': item.get('time_to_retire'),
            u'type': item.get('type')
        })

    def _response_from_item(self, item):
        return remove_nones_from_dict({
            u'ipAddress': item.get(u'ipAddress'),
            u'timeToRetire': item.get(u'timeToRetire'),
            u'type': item.get(u'type')
        })


class InstanceReplicaConfiguration(object):
    def __init__(self, request, module):
        self.module = module
        if request:
            self.request = request
        else:
            self.request = {}

    def to_request(self):
        return remove_nones_from_dict({
            u'failoverTarget': self.request.get('failover_target'),
            u'mysqlReplicaConfiguration': InstanceMysqlReplicaConfiguration(self.request.get('mysql_replica_configuration', {}), self.module).to_request(),
            u'replicaNames': self.request.get('replica_names'),
            u'serviceAccountEmailAddress': self.request.get('service_account_email_address')
        })

    def from_response(self):
        return remove_nones_from_dict({
            u'failoverTarget': self.request.get(u'failoverTarget'),
            u'mysqlReplicaConfiguration': InstanceMysqlReplicaConfiguration(self.request.get(u'mysqlReplicaConfiguration', {}), self.module).from_response(),
            u'replicaNames': self.request.get(u'replicaNames'),
            u'serviceAccountEmailAddress': self.request.get(u'serviceAccountEmailAddress')
        })


class InstanceMysqlReplicaConfiguration(object):
    def __init__(self, request, module):
        self.module = module
        if request:
            self.request = request
        else:
            self.request = {}

    def to_request(self):
        return remove_nones_from_dict({
            u'caCertificate': self.request.get('ca_certificate'),
            u'clientCertificate': self.request.get('client_certificate'),
            u'clientKey': self.request.get('client_key'),
            u'connectRetryInterval': self.request.get('connect_retry_interval'),
            u'dumpFilePath': self.request.get('dump_file_path'),
            u'masterHeartbeatPeriod': self.request.get('master_heartbeat_period'),
            u'password': self.request.get('password'),
            u'sslCipher': self.request.get('ssl_cipher'),
            u'username': self.request.get('username'),
            u'verifyServerCertificate': self.request.get('verify_server_certificate')
        })

    def from_response(self):
        return remove_nones_from_dict({
            u'caCertificate': self.request.get(u'caCertificate'),
            u'clientCertificate': self.request.get(u'clientCertificate'),
            u'clientKey': self.request.get(u'clientKey'),
            u'connectRetryInterval': self.request.get(u'connectRetryInterval'),
            u'dumpFilePath': self.request.get(u'dumpFilePath'),
            u'masterHeartbeatPeriod': self.request.get(u'masterHeartbeatPeriod'),
            u'password': self.request.get(u'password'),
            u'sslCipher': self.request.get(u'sslCipher'),
            u'username': self.request.get(u'username'),
            u'verifyServerCertificate': self.request.get(u'verifyServerCertificate')
        })


class InstanceSettings(object):
    def __init__(self, request, module):
        self.module = module
        if request:
            self.request = request
        else:
            self.request = {}

    def to_request(self):
        return remove_nones_from_dict({
            u'ipConfiguration': InstanceIpConfiguration(self.request.get('ip_configuration', {}), self.module).to_request(),
            u'tier': self.request.get('tier')
        })

    def from_response(self):
        return remove_nones_from_dict({
            u'ipConfiguration': InstanceIpConfiguration(self.request.get(u'ipConfiguration', {}), self.module).from_response(),
            u'tier': self.request.get(u'tier')
        })


class InstanceIpConfiguration(object):
    def __init__(self, request, module):
        self.module = module
        if request:
            self.request = request
        else:
            self.request = {}

    def to_request(self):
        return remove_nones_from_dict({
            u'ipv4Enabled': self.request.get('ipv4_enabled'),
            u'authorizedNetworks': InstanceAuthorizedNetworksArray(self.request.get('authorized_networks', []), self.module).to_request(),
            u'requireSsl': self.request.get('require_ssl')
        })

    def from_response(self):
        return remove_nones_from_dict({
            u'ipv4Enabled': self.request.get(u'ipv4Enabled'),
            u'authorizedNetworks': InstanceAuthorizedNetworksArray(self.request.get(u'authorizedNetworks', []), self.module).from_response(),
            u'requireSsl': self.request.get(u'requireSsl')
        })


class InstanceAuthorizedNetworksArray(object):
    def __init__(self, request, module):
        self.module = module
        if request:
            self.request = request
        else:
            self.request = []

    def to_request(self):
        items = []
        for item in self.request:
            items.append(self._request_for_item(item))
        return items

    def from_response(self):
        items = []
        for item in self.request:
            items.append(self._response_from_item(item))
        return items

    def _request_for_item(self, item):
        return remove_nones_from_dict({
            u'expirationTime': item.get('expiration_time'),
            u'name': item.get('name'),
            u'value': item.get('value')
        })

    def _response_from_item(self, item):
        return remove_nones_from_dict({
            u'expirationTime': item.get(u'expirationTime'),
            u'name': item.get(u'name'),
            u'value': item.get(u'value')
        })


if __name__ == '__main__':
    main()



#!/usr/bin/python

# (c) 2018, NetApp, Inc
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
from __future__ import absolute_import, division, print_function
__metaclass__ = type

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}


DOCUMENTATION = '''
author: NetApp Ansible Team (ng-ansibleteam@netapp.com)
description:
  - Update ONTAP software
extends_documentation_fragment:
  - netapp.na_ontap
module: na_ontap_software_update
options:
  state:
    choices: ['present', 'absent']
    description:
      - Whether the specified ONTAP package should update or not.
    default: present
  node:
    description:
      - List of nodes to be updated, the nodes have to be a part of a HA Pair.
  package_version:
    required: true
    description:
      - Specifies the package version.
  package_url:
    required: true
    description:
      - Specifies the package URL.
  ignore_validation_warning:
    description:
      - Allows the update to continue if warnings are encountered during the validation phase.
    default: False
    type: bool
short_description: NetApp ONTAP Update Software
version_added: "2.7"
'''

EXAMPLES = """

    - name: ONTAP software update
      na_ontap_software_update:
        state: present
        node: laurentn-vsim1
        package_url: "{{ url }}"
        package_version: "{{ version_name }}"
        ignore_validation_warning: True
        hostname: "{{ netapp_hostname }}"
        username: "{{ netapp_username }}"
        password: "{{ netapp_password }}"
"""

RETURN = """
"""

import traceback
from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils._text import to_native
import ansible.module_utils.netapp as netapp_utils
from ansible.module_utils.netapp_module import NetAppModule
import time

HAS_NETAPP_LIB = netapp_utils.has_netapp_lib()


class NetAppONTAPSoftwareUpdate(object):
    """
    Class with ONTAP software update methods
    """

    def __init__(self):

        self.argument_spec = netapp_utils.na_ontap_host_argument_spec()
        self.argument_spec.update(dict(
            state=dict(required=False, type='str', choices=['present', 'absent'], default='present'),
            node=dict(required=False, type='list'),
            package_version=dict(required=True, type='str'),
            package_url=dict(required=True, type='str'),
            ignore_validation_warning=dict(required=False, type='bool', default=False)
        ))

        self.module = AnsibleModule(
            argument_spec=self.argument_spec,
            supports_check_mode=True
        )

        self.na_helper = NetAppModule()
        self.parameters = self.na_helper.set_parameters(self.module.params)

        if HAS_NETAPP_LIB is False:
            self.module.fail_json(msg="the python NetApp-Lib module is required")
        else:
            self.server = netapp_utils.setup_na_ontap_zapi(module=self.module)

    def cluster_image_get_iter(self):
        """
        Compose NaElement object to query current version
        :return: NaElement object for cluster-image-get-iter with query
        """
        cluster_image_get = netapp_utils.zapi.NaElement('cluster-image-get-iter')
        query = netapp_utils.zapi.NaElement('query')
        cluster_image_info = netapp_utils.zapi.NaElement('cluster-image-info')
        query.add_child_elem(cluster_image_info)
        cluster_image_get.add_child_elem(query)
        return cluster_image_get

    def cluster_image_get(self):
        """
        Get current cluster image info
        :return: True if query successful, else return None
        """
        cluster_image_get_iter = self.cluster_image_get_iter()
        try:
            result = self.server.invoke_successfully(cluster_image_get_iter, enable_tunneling=True)
        except netapp_utils.zapi.NaApiError as error:
            self.module.fail_json(msg='Error fetching cluster image details for %s: %s'
                                      % (self.parameters['node'], to_native(error)),
                                  exception=traceback.format_exc())
        # return cluster image details
        if result.get_child_by_name('num-records') and \
                int(result.get_child_content('num-records')) > 0:
            return True
        return None

    def cluster_image_update(self):
        """
        Update current cluster image
        :return: None
        """
        cluster_update_info = netapp_utils.zapi.NaElement('cluster-image-update')
        cluster_update_info.add_new_child('package-version', self.parameters['package_version'])
        cluster_update_info.add_new_child('ignore-validation-warning',
                                          str(self.parameters['ignore_validation_warning']))
        if self.parameters.get('nodes'):
            cluster_nodes = netapp_utils.zapi.NaElement('nodes')
            for node in self.parameters['node']:
                cluster_nodes.add_new_child('node-name', node)
            cluster_update_info.add_child_elem(cluster_nodes)
        try:
            self.server.invoke_successfully(cluster_update_info, enable_tunneling=True)
        except netapp_utils.zapi.NaApiError as error:
            self.module.fail_json(msg='Error updating cluster image for %s: %s'
                                      % (self.parameters['package_version'], to_native(error)),
                                  exception=traceback.format_exc())

    def cluster_image_package_download(self):
        """
        Get current cluster image package download
        :return: True if package already exists, else return False
        """
        cluster_image_package_download_info = netapp_utils.zapi.NaElement('cluster-image-package-download')
        cluster_image_package_download_info.add_new_child('package-url', self.parameters['package_url'])
        try:
            self.server.invoke_successfully(cluster_image_package_download_info, enable_tunneling=True)
        except netapp_utils.zapi.NaApiError as error:
            # Error 18408 denotes Package image with the same name already exists
            if to_native(error.code) == "18408":
                return True
            else:
                self.module.fail_json(msg='Error downloading cluster image package for %s: %s'
                                          % (self.parameters['package_url'], to_native(error)),
                                      exception=traceback.format_exc())
        return False

    def cluster_image_package_download_progress(self):
        """
        Get current cluster image package download progress
        :return: Dictionary of cluster image download progress if query successful, else return None
        """
        cluster_image_package_download_progress_info = netapp_utils.zapi.\
            NaElement('cluster-image-get-download-progress')
        try:
            result = self.server.invoke_successfully(
                cluster_image_package_download_progress_info, enable_tunneling=True)
        except netapp_utils.zapi.NaApiError as error:
            self.module.fail_json(msg='Error fetching cluster image package download progress for %s: %s'
                                      % (self.parameters['package_url'], to_native(error)),
                                  exception=traceback.format_exc())
        # return cluster image download progress details
        cluster_download_progress_info = dict()
        if result.get_child_by_name('progress-status'):
            cluster_download_progress_info['progress_status'] = result.get_child_content('progress-status')
            cluster_download_progress_info['progress_details'] = result.get_child_content('progress-details')
            cluster_download_progress_info['failure_reason'] = result.get_child_content('failure-reason')
            return cluster_download_progress_info
        return None

    def apply(self):
        """
        Apply action to update ONTAP software
        """
        changed = False
        current = self.cluster_image_get()
        results = netapp_utils.get_cserver(self.server)
        cserver = netapp_utils.setup_ontap_zapi(module=self.module, vserver=results)
        netapp_utils.ems_log_event("na_ontap_software_update", cserver)
        if self.parameters.get('state') == 'present' and current:
            package_exists = self.cluster_image_package_download()
            if package_exists is False:
                cluster_download_progress = self.cluster_image_package_download_progress()
                while cluster_download_progress.get('progress_status') == 'async_pkg_get_phase_running':
                    time.sleep(5)
                    cluster_download_progress = self.cluster_image_package_download_progress()
                if cluster_download_progress.get('progress_status') == 'async_pkg_get_phase_complete':
                    self.cluster_image_update()
                    changed = True
                else:
                    self.module.fail_json(msg='Error downloading package: %s'
                                              % (cluster_download_progress['failure_reason']))
            else:
                self.cluster_image_update()
                changed = True
        self.module.exit_json(changed=changed)


def main():
    """Execute action"""
    community_obj = NetAppONTAPSoftwareUpdate()
    community_obj.apply()


if __name__ == '__main__':
    main()



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.


class ModuleDocFragment(object):

    # Standard openstack documentation fragment
    DOCUMENTATION = '''
options:
  cloud:
    description:
      - Named cloud or cloud config to operate against.
        If I(cloud) is a string, it references a named cloud config as defined
        in an OpenStack clouds.yaml file. Provides default values for I(auth)
        and I(auth_type). This parameter is not needed if I(auth) is provided
        or if OpenStack OS_* environment variables are present.
        If I(cloud) is a dict, it contains a complete cloud configuration like
        would be in a section of clouds.yaml.
    required: false
  auth:
    description:
      - Dictionary containing auth information as needed by the cloud's auth
        plugin strategy. For the default I(password) plugin, this would contain
        I(auth_url), I(username), I(password), I(project_name) and any
        information about domains if the cloud supports them. For other plugins,
        this param will need to contain whatever parameters that auth plugin
        requires. This parameter is not needed if a named cloud is provided or
        OpenStack OS_* environment variables are present.
    required: false
  auth_type:
    description:
      - Name of the auth plugin to use. If the cloud uses something other than
        password authentication, the name of the plugin should be indicated here
        and the contents of the I(auth) parameter should be updated accordingly.
    required: false
  region_name:
    description:
      - Name of the region.
    required: false
  wait:
    description:
      - Should ansible wait until the requested resource is complete.
    type: bool
    required: false
    default: true
  timeout:
    description:
      - How long should ansible wait for the requested resource.
    required: false
    default: 180
  api_timeout:
    description:
      - How long should the socket layer wait before timing out for API calls.
        If this is omitted, nothing will be passed to the requests library.
    required: false
  verify:
    description:
      - Whether or not SSL API requests should be verified. Before 2.3 this defaulted to True.
    type: bool
    required: false
    aliases: ['validate_certs']
  cacert:
    description:
      - A path to a CA Cert bundle that can be used as part of verifying
        SSL API requests.
    required: false
  cert:
    description:
      - A path to a client certificate to use as part of the SSL transaction.
    required: false
  key:
    description:
      - A path to a client key to use as part of the SSL transaction.
    required: false
  interface:
    description:
        - Endpoint URL type to fetch from the service catalog.
    choices: [public, internal, admin]
    required: false
    default: public
    aliases: ['endpoint_type']
    version_added: "2.3"
requirements:
  - python >= 2.7
  - openstacksdk
notes:
  - The standard OpenStack environment variables, such as C(OS_USERNAME)
    may be used instead of providing explicit values.
  - Auth information is driven by os-client-config, which means that values
    can come from a yaml config file in /etc/ansible/openstack.yaml,
    /etc/openstack/clouds.yaml or ~/.config/openstack/clouds.yaml, then from
    standard environment variables, then finally by explicit parameters in
    plays. More information can be found at
    U(http://docs.openstack.org/developer/os-client-config)
'''



# (c) 2012, Jan-Piet Mens <jpmens () gmail.com>
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
#

# Make coding more python3-ish
from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

from collections import MutableMapping, MutableSet, MutableSequence

from ansible.errors import AnsibleError, AnsibleAssertionError
from ansible.module_utils.six import string_types
from ansible.module_utils._text import to_native
from ansible.parsing.plugin_docs import read_docstring, read_docstub
from ansible.parsing.yaml.loader import AnsibleLoader

try:
    from __main__ import display
except ImportError:
    from ansible.utils.display import Display
    display = Display()


# modules that are ok that they do not have documentation strings
BLACKLIST = {
    'MODULE': frozenset(('async_wrapper',)),
    'CACHE': frozenset(('base',)),
}


def merge_fragment(target, source):

    for key, value in source.items():
        if key in target:
            # assumes both structures have same type
            if isinstance(target[key], MutableMapping):
                value.update(target[key])
            elif isinstance(target[key], MutableSet):
                value.add(target[key])
            elif isinstance(target[key], MutableSequence):
                value = sorted(frozenset(value + target[key]))
            else:
                raise Exception("Attempt to extend a documentation fragement, invalid type for %s" % key)
        target[key] = value


def add_fragments(doc, filename, fragment_loader):

    fragments = doc.pop('extends_documentation_fragment', [])

    if isinstance(fragments, string_types):
        fragments = [fragments]

    # Allow the module to specify a var other than DOCUMENTATION
    # to pull the fragment from, using dot notation as a separator
    for fragment_slug in fragments:
        fragment_slug = fragment_slug.lower()
        if '.' in fragment_slug:
            fragment_name, fragment_var = fragment_slug.split('.', 1)
            fragment_var = fragment_var.upper()
        else:
            fragment_name, fragment_var = fragment_slug, 'DOCUMENTATION'

        fragment_class = fragment_loader.get(fragment_name)
        if fragment_class is None:
            raise AnsibleAssertionError('fragment_class is None')

        fragment_yaml = getattr(fragment_class, fragment_var, '{}')
        fragment = AnsibleLoader(fragment_yaml, file_name=filename).get_single_data()

        if 'notes' in fragment:
            notes = fragment.pop('notes')
            if notes:
                if 'notes' not in doc:
                    doc['notes'] = []
                doc['notes'].extend(notes)

        if 'seealso' in fragment:
            seealso = fragment.pop('seealso')
            if seealso:
                if 'seealso' not in doc:
                    doc['seealso'] = []
                doc['seealso'].extend(seealso)

        if 'options' not in fragment:
            raise Exception("missing options in fragment (%s), possibly misformatted?: %s" % (fragment_name, filename))

        # ensure options themselves are directly merged
        if 'options' in doc:
            try:
                merge_fragment(doc['options'], fragment.pop('options'))
            except Exception as e:
                raise AnsibleError("%s options (%s) of unknown type: %s" % (to_native(e), fragment_name, filename))
        else:
            doc['options'] = fragment.pop('options')

        # merge rest of the sections
        try:
            merge_fragment(doc, fragment)
        except Exception as e:
            raise AnsibleError("%s (%s) of unknown type: %s" % (to_native(e), fragment_name, filename))


def get_docstring(filename, fragment_loader, verbose=False, ignore_errors=False):
    """
    DOCUMENTATION can be extended using documentation fragments loaded by the PluginLoader from the module_docs_fragments directory.
    """

    data = read_docstring(filename, verbose=verbose, ignore_errors=ignore_errors)

    # add fragments to documentation
    if data.get('doc', False):
        add_fragments(data['doc'], filename, fragment_loader=fragment_loader)

    return data['doc'], data['plainexamples'], data['returndocs'], data['metadata']



#!/usr/bin/python
#
# (c) 2015 Peter Sprygada, <psprygada@ansible.com>
# Copyright (c) 2016 Dell Inc.
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}


DOCUMENTATION = """
---
module: dellos6_facts
version_added: "2.2"
author: "Abirami N(@abirami-n)"
short_description: Collect facts from remote devices running Dell EMC Networking OS6
description:
  - Collects a base set of device facts from a remote device that
    is running OS6.  This module prepends all of the
    base network fact keys with C(ansible_net_<fact>).  The facts
    module will always collect a base set of facts from the device
    and can enable or disable collection of additional facts.
extends_documentation_fragment: dellos6
options:
  gather_subset:
    description:
      - When supplied, this argument will restrict the facts collected
        to a given subset.  Possible values for this argument include
        all, hardware, config, and interfaces. Can specify a list of
        values to include a larger subset.  Values can also be used
        with an initial C(M(!)) to specify that a specific subset should
        not be collected.
    default: [ '!config' ]
"""

EXAMPLES = """
# Collect all facts from the device
- dellos6_facts:
    gather_subset: all

# Collect only the config and default facts
- dellos6_facts:
    gather_subset:
      - config

# Do not collect hardware facts
- dellos6_facts:
    gather_subset:
      - "!interfaces"
"""

RETURN = """
ansible_net_gather_subset:
  description: The list of fact subsets collected from the device.
  returned: always.
  type: list

# default
ansible_net_model:
  description: The model name returned from the device.
  returned: always.
  type: str
ansible_net_serialnum:
  description: The serial number of the remote device.
  returned: always.
  type: str
ansible_net_version:
  description: The operating system version running on the remote device.
  returned: always.
  type: str
ansible_net_hostname:
  description: The configured hostname of the device.
  returned: always.
  type: string
ansible_net_image:
  description: The image file that the device is running.
  returned: always
  type: string

# hardware
ansible_net_memfree_mb:
  description: The available free memory on the remote device in MB.
  returned: When hardware is configured.
  type: int
ansible_net_memtotal_mb:
  description: The total memory on the remote device in MB.
  returned: When hardware is configured.
  type: int

# config
ansible_net_config:
  description: The current active config from the device.
  returned: When config is configured.
  type: str

# interfaces
ansible_net_interfaces:
  description: A hash of all interfaces running on the system.
  returned: When interfaces is configured.
  type: dict
ansible_net_neighbors:
  description: The list of LLDP neighbors from the remote device.
  returned: When interfaces is configured.
  type: dict

"""
import re

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.network.dellos6.dellos6 import run_commands
from ansible.module_utils.network.dellos6.dellos6 import dellos6_argument_spec, check_args
from ansible.module_utils.six import iteritems


class FactsBase(object):

    COMMANDS = list()

    def __init__(self, module):
        self.module = module
        self.facts = dict()
        self.responses = None

    def populate(self):
        self.responses = run_commands(self.module, self.COMMANDS, check_rc=False)

    def run(self, cmd):
        return run_commands(self.module, cmd, check_rc=False)


class Default(FactsBase):

    COMMANDS = [
        'show version',
        'show running-config | include hostname'
    ]

    def populate(self):
        super(Default, self).populate()
        data = self.responses[0]
        self.facts['version'] = self.parse_version(data)
        self.facts['serialnum'] = self.parse_serialnum(data)
        self.facts['model'] = self.parse_model(data)
        self.facts['image'] = self.parse_image(data)
        hdata = self.responses[1]
        self.facts['hostname'] = self.parse_hostname(hdata)

    def parse_version(self, data):
        match = re.search(r'HW Version(.+)\s(\d+)', data)
        if match:
            return match.group(2)

    def parse_hostname(self, data):
        match = re.search(r'\S+\s(\S+)', data, re.M)
        if match:
            return match.group(1)

    def parse_model(self, data):
        match = re.search(r'System Model ID(.+)\s([A-Z0-9]*)\n', data, re.M)
        if match:
            return match.group(2)

    def parse_image(self, data):
        match = re.search(r'Image File(.+)\s([A-Z0-9a-z_.]*)\n', data)
        if match:
            return match.group(2)

    def parse_serialnum(self, data):
        match = re.search(r'Serial Number(.+)\s([A-Z0-9]*)\n', data)
        if match:
            return match.group(2)


class Hardware(FactsBase):

    COMMANDS = [
        'show memory cpu'
    ]

    def populate(self):
        super(Hardware, self).populate()
        data = self.responses[0]
        match = re.findall(r'\s(\d+)\s', data)
        if match:
            self.facts['memtotal_mb'] = int(match[0]) // 1024
            self.facts['memfree_mb'] = int(match[1]) // 1024


class Config(FactsBase):

    COMMANDS = ['show running-config']

    def populate(self):
        super(Config, self).populate()
        self.facts['config'] = self.responses[0]


class Interfaces(FactsBase):
    COMMANDS = [
        'show interfaces',
        'show interfaces status',
        'show interfaces transceiver properties',
        'show ip int',
        'show lldp',
        'show lldp remote-device all'
    ]

    def populate(self):
        vlan_info = dict()
        super(Interfaces, self).populate()
        data = self.responses[0]
        interfaces = self.parse_interfaces(data)
        desc = self.responses[1]
        properties = self.responses[2]
        vlan = self.responses[3]
        vlan_info = self.parse_vlan(vlan)
        self.facts['interfaces'] = self.populate_interfaces(interfaces, desc, properties)
        self.facts['interfaces'].update(vlan_info)
        if 'LLDP is not enabled' not in self.responses[4]:
            neighbors = self.responses[5]
            self.facts['neighbors'] = self.parse_neighbors(neighbors)

    def parse_vlan(self, vlan):
        facts = dict()
        vlan_info, vlan_info_next = vlan.split('----------   -----   --------------- --------------- -------')
        for en in vlan_info_next.splitlines():
            if en == '':
                continue
            match = re.search(r'^(\S+)\s+(\S+)\s+(\S+)', en)
            intf = match.group(1)
            if intf not in facts:
                facts[intf] = list()
            fact = dict()
            matc = re.search(r'^([\w+\s\d]*)\s+(\S+)\s+(\S+)', en)
            fact['address'] = matc.group(2)
            fact['masklen'] = matc.group(3)
            facts[intf].append(fact)
        return facts

    def populate_interfaces(self, interfaces, desc, properties):
        facts = dict()
        for key, value in interfaces.items():
            intf = dict()
            intf['description'] = self.parse_description(key, desc)
            intf['macaddress'] = self.parse_macaddress(value)
            intf['mtu'] = self.parse_mtu(value)
            intf['bandwidth'] = self.parse_bandwidth(value)
            intf['mediatype'] = self.parse_mediatype(key, properties)
            intf['duplex'] = self.parse_duplex(value)
            intf['lineprotocol'] = self.parse_lineprotocol(value)
            intf['operstatus'] = self.parse_operstatus(value)
            intf['type'] = self.parse_type(key, properties)
            facts[key] = intf
        return facts

    def parse_neighbors(self, neighbors):
        facts = dict()
        neighbor, neighbor_next = neighbors.split('--------- ------- ------------------- ----------------- -----------------')
        for en in neighbor_next.splitlines():
            if en == '':
                continue
            intf = self.parse_lldp_intf(en.split()[0])
            if intf not in facts:
                facts[intf] = list()
            fact = dict()
            fact['port'] = self.parse_lldp_port(en.split()[3])
            if (len(en.split()) > 4):
                fact['host'] = self.parse_lldp_host(en.split()[4])
            else:
                fact['host'] = "Null"
            facts[intf].append(fact)

        return facts

    def parse_interfaces(self, data):
        parsed = dict()
        for line in data.split('\n'):
            if len(line) == 0:
                continue
            else:
                match = re.match(r'Interface Name(.+)\s([A-Za-z0-9/]*)', line)
                if match:
                    key = match.group(2)
                    parsed[key] = line
                else:
                    parsed[key] += '\n%s' % line
        return parsed

    def parse_description(self, key, desc):
        desc, desc_next = desc.split('--------- --------------- ------ ------- ---- ------ ----- -- -------------------')
        if desc_next.find('Oob') > 0:
            desc_val, desc_info = desc_next.split('Oob')
        elif desc_next.find('Port') > 0:
            desc_val, desc_info = desc_next.split('Port')
        for en in desc_val.splitlines():
            if key in en:
                match = re.search(r'^(\S+)\s+(\S+)', en)
                if match.group(2) in ['Full', 'N/A']:
                    return "Null"
                else:
                    return match.group(2)

    def parse_macaddress(self, data):
        match = re.search(r'Burned MAC Address(.+)\s([A-Z0-9.]*)\n', data)
        if match:
            return match.group(2)

    def parse_mtu(self, data):
        match = re.search(r'MTU Size(.+)\s(\d+)\n', data)
        if match:
            return int(match.group(2))

    def parse_bandwidth(self, data):
        match = re.search(r'Port Speed\s*[:\s\.]+\s(\d+)\n', data)
        if match:
            return int(match.group(1))

    def parse_duplex(self, data):
        match = re.search(r'Port Mode\s([A-Za-z]*)(.+)\s([A-Za-z/]*)\n', data)
        if match:
            return match.group(3)

    def parse_mediatype(self, key, properties):
        mediatype, mediatype_next = properties.split('--------- ------- --------------------- --------------------- --------------')
        flag = 1
        for en in mediatype_next.splitlines():
            if key in en:
                flag = 0
                match = re.search(r'^(\S+)\s+(\S+)\s+(\S+)', en)
                if match:
                    strval = match.group(3)
                    return strval
        if flag == 1:
            return "null"

    def parse_type(self, key, properties):
        type_val, type_val_next = properties.split('--------- ------- --------------------- --------------------- --------------')
        flag = 1
        for en in type_val_next.splitlines():
            if key in en:
                flag = 0
                match = re.search(r'^(\S+)\s+(\S+)\s+(\S+)', en)
                if match:
                    strval = match.group(2)
                    return strval
        if flag == 1:
            return "null"

    def parse_lineprotocol(self, data):
        data = data.splitlines()
        for d in data:
            match = re.search(r'^Link Status\s*[:\s\.]+\s(\S+)', d)
            if match:
                return match.group(1)

    def parse_operstatus(self, data):
        data = data.splitlines()
        for d in data:
            match = re.search(r'^Link Status\s*[:\s\.]+\s(\S+)', d)
            if match:
                return match.group(1)

    def parse_lldp_intf(self, data):
        match = re.search(r'^([A-Za-z0-9/]*)', data)
        if match:
            return match.group(1)

    def parse_lldp_host(self, data):
        match = re.search(r'^([A-Za-z0-9-]*)', data)
        if match:
            return match.group(1)

    def parse_lldp_port(self, data):
        match = re.search(r'^([A-Za-z0-9/]*)', data)
        if match:
            return match.group(1)


FACT_SUBSETS = dict(
    default=Default,
    hardware=Hardware,
    interfaces=Interfaces,
    config=Config,
)

VALID_SUBSETS = frozenset(FACT_SUBSETS.keys())


def main():
    """main entry point for module execution
    """
    argument_spec = dict(
        gather_subset=dict(default=['!config'], type='list')
    )

    argument_spec.update(dellos6_argument_spec)

    module = AnsibleModule(argument_spec=argument_spec,
                           supports_check_mode=True)

    gather_subset = module.params['gather_subset']

    runable_subsets = set()
    exclude_subsets = set()

    for subset in gather_subset:
        if subset == 'all':
            runable_subsets.update(VALID_SUBSETS)
            continue

        if subset.startswith('!'):
            subset = subset[1:]
            if subset == 'all':
                exclude_subsets.update(VALID_SUBSETS)
                continue
            exclude = True
        else:
            exclude = False

        if subset not in VALID_SUBSETS:
            module.fail_json(msg='Bad subset')

        if exclude:
            exclude_subsets.add(subset)
        else:
            runable_subsets.add(subset)

    if not runable_subsets:
        runable_subsets.update(VALID_SUBSETS)

    runable_subsets.difference_update(exclude_subsets)
    runable_subsets.add('default')

    facts = dict()
    facts['gather_subset'] = list(runable_subsets)

    instances = list()
    for key in runable_subsets:
        instances.append(FACT_SUBSETS[key](module))

    for inst in instances:
        inst.populate()
        facts.update(inst.facts)

    ansible_facts = dict()
    for key, value in iteritems(facts):
        key = 'ansible_net_%s' % key
        ansible_facts[key] = value

    warnings = list()
    check_args(module, warnings)

    module.exit_json(ansible_facts=ansible_facts, warnings=warnings)


if __name__ == '__main__':
    main()



#!/usr/bin/python
# (c) 2018, NetApp, Inc
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

"""
Element Software Backup Manager
"""
from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'certified'}

DOCUMENTATION = '''

module: na_elementsw_backup

short_description: NetApp Element Software Create Backups
extends_documentation_fragment:
    - netapp.solidfire
version_added: '2.7'
author: NetApp Ansible Team (ng-ansibleteam@netapp.com)
description:
- Create backup

options:

    src_hostname:
        description:
        - hostname for the backup source cluster
        required: true
        aliases:
        - hostname

    src_username:
        description:
        - username for the backup source cluster
        required: true
        aliases:
        - username
        - user

    src_password:
        description:
        - password for the backup source cluster
        required: true
        aliases:
        - password
        - pass

    src_volume_id:
        description:
        - ID of the backup source volume.
        required: true
        aliases:
        - volume_id

    dest_hostname:
        description:
        - hostname for the backup source cluster
        - will be set equal to src_hostname if not specified
        required: false

    dest_username:
        description:
        - username for the backup destination cluster
        - will be set equal to src_username if not specified
        required: false

    dest_password:
        description:
        - password for the backup destination cluster
        - will be set equal to src_password if not specified
        required: false

    dest_volume_id:
        description:
        - ID of the backup destination volume
        required: true

    format:
        description:
        - Backup format to use
        choices: ['native','uncompressed']
        required: false
        default: 'native'

    script:
        description:
        - the backup script to be executed
        required: false

    script_parameters:
        description:
        - the backup script parameters
        required: false

'''

EXAMPLES = """
na_elementsw_backup:
  src_hostname: "{{ source_cluster_hostname }}"
  src_username: "{{ source_cluster_username }}"
  src_password: "{{ source_cluster_password }}"
  src_volume_id: 1
  dest_hostname: "{{ destination_cluster_hostname }}"
  dest_username: "{{ destination_cluster_username }}"
  dest_password: "{{ destination_cluster_password }}"
  dest_volume_id: 3
  format: native
"""

RETURN = """

"""

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils._text import to_native
import ansible.module_utils.netapp as netapp_utils
from ansible.module_utils.netapp_elementsw_module import NaElementSWModule
import time

HAS_SF_SDK = netapp_utils.has_sf_sdk()
try:
    import solidfire.common
except:
    HAS_SF_SDK = False


class ElementSWBackup(object):
    ''' class to handle backup operations '''

    def __init__(self):
        """
            Setup Ansible parameters and SolidFire connection
        """
        self.argument_spec = {}
        self.argument_spec.update(dict(

            src_hostname=dict(aliases=['hostname'], required=True, type='str'),
            src_username=dict(aliases=['username', 'user'], required=True, type='str'),
            src_password=dict(aliases=['password', 'pass'], required=True, type='str', no_log=True),
            src_volume_id=dict(aliases=['volume_id'], required=True, type='str'),
            dest_hostname=dict(required=False, type='str'),
            dest_username=dict(required=False, type='str'),
            dest_password=dict(required=False, type='str', no_log=True),
            dest_volume_id=dict(required=True, type='str'),
            format=dict(required=False, choices=['native', 'uncompressed'], default='native'),
            script=dict(required=False, type='str'),
            script_parameters=dict(required=False, type='dict')


        ))

        self.module = AnsibleModule(
            argument_spec=self.argument_spec,
            required_together=[['script', 'script_parameters']],
            supports_check_mode=True
        )
        if HAS_SF_SDK is False:
            self.module.fail_json(msg="Unable to import the SolidFire Python SDK")

        # If destination cluster details are not specified , set the destination to be the same as the source
        if self.module.params["dest_hostname"] is None:
            self.module.params["dest_hostname"] = self.module.params["src_hostname"]
        if self.module.params["dest_username"] is None:
            self.module.params["dest_username"] = self.module.params["src_username"]
        if self.module.params["dest_password"] is None:
            self.module.params["dest_password"] = self.module.params["src_password"]

        params = self.module.params

        # establish a connection to both source and destination sf clusters

        self.module.params["username"] = params["src_username"]
        self.module.params["password"] = params["src_password"]
        self.module.params["hostname"] = params["src_hostname"]
        self.src_connection = netapp_utils.create_sf_connection(self.module)
        self.module.params["username"] = params["dest_username"]
        self.module.params["password"] = params["dest_password"]
        self.module.params["hostname"] = params["dest_hostname"]
        self.dest_connection = netapp_utils.create_sf_connection(self.module)

        self.elementsw_helper = NaElementSWModule(self.sfe)

        # add telemetry attributes
        self.attributes = self.elementsw_helper.set_element_attributes(source='na_elementsw_backup')

    def apply(self):
        """
            Apply backup creation logic
        """
        self.create_backup()
        self.module.exit_json(changed=True)

    def create_backup(self):
        """
            Create backup
        """

        # Start volume write on destination cluster

        try:
            write_obj = self.dest_connection.start_bulk_volume_write(volume_id=self.module.params["dest_volume_id"],
                                                                     format=self.module.params["format"],
                                                                     attributes=self.attributes)
            write_key = write_obj.key
        except solidfire.common.ApiServerError as err:
            self.module.fail_json(msg="Error starting bulk write on destination cluster", exception=to_native(err))

        # Set script parameters if not passed by user
        # These parameters are equivalent to the options used when a backup is executed via the GUI

        if self.module.params["script"] is None and self.module.params["script_parameters"] is None:

            self.module.params["script"] = 'bv_internal.py'
            self.module.params["script_parameters"] = {"write": {
                "mvip": self.module.params["dest_hostname"],
                "username": self.module.params["dest_username"],
                "password": self.module.params["dest_password"],
                "key": write_key,
                "endpoint": "solidfire",
                "format": self.module.params["format"]},
                "range": {"lba": 0, "blocks": 244224}}

        # Start volume read on source cluster

        try:
            read_obj = self.src_connection.start_bulk_volume_read(self.module.params["src_volume_id"],
                                                                  self.module.params["format"],
                                                                  script=self.module.params["script"],
                                                                  script_parameters=self.module.params["script_parameters"],
                                                                  attributes=self.attributes)
        except solidfire.common.ApiServerError as err:
            self.module.fail_json(msg="Error starting bulk read on source cluster", exception=to_native(err))

        # Poll job status until it has completed
        # SF will automatically timeout if the job is not successful after certain amount of time

        completed = False
        while completed is not True:
            # Sleep between polling iterations to reduce api load
            time.sleep(2)
            try:
                result = self.src_connection.get_async_result(read_obj.async_handle, True)
            except solidfire.common.ApiServerError as err:
                self.module.fail_json(msg="Unable to check backup job status", exception=to_native(err))

            if result["status"] != 'running':
                completed = True
        if 'error' in result:
            self.module.fail_json(msg=result['error']['message'])


def main():
    """ Run backup operation"""
    vol_obj = ElementSWBackup()
    vol_obj.apply()


if __name__ == '__main__':
    main()
